/*
    File: AfhFMeditor.c
    C source file for the default (Lite) source editor for the
    AFH Filter Meister Photoshop Plug-in Filter.

 	Copyright 2000-2014 AFH Systems.
    Derived from AFHFMUI.c.

  Derived from Photoshop SDK source files which are:

    Copyright 1991-95 by Adobe Systems, Inc.

  FilterMeister is free software. You can redistribute and modify it
  under the terms of the MPL2 License.
  See: https://www.mozilla.org/en-US/MPL/2.0/
  Any changes to the FilterMeister files have to be published or sent
  to harry@thepluginsite.com. The filter plugin code itself does not
  need to be disclosed.


	16JUL13  A. Hunter  1. Modify for WIN64.
	16NOV13  A. Hunter  1. Add ???
	06FEB14	 A. Hunter	1. Add PREPROCESS and X64GEN controls.

*/

#ifndef STANDALONE

#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include <windowsx.h>       // For SetDlgMsgResult()
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>         /* offsetof */
#include <process.h>        // _beginthread
#include <time.h>
#include <assert.h>
#include "PITypes.h"
#include "PIGeneral.h"
#include "PIFilter.h"
#include "PIAbout.h"
#include "PIProperties.h"
#include "PIUtilities.h"

#include "AfhFM.h"
#include "AfhFMcontext.h"
#include "mdp.h"            //parser flags

#include "md386def.hc"      //MD386__SYNTAX

#include "resource.h"

#include "WinUtil.h"    // Message crackers and stuff

#ifndef USE_EDIT_CONTROL
#include "RichEdit.h"   //Rich Edit Control definitions
#endif //USE_EDIT_CONTROL


HWND hEdit1;
HWND hEdit2;
HWND hEditorDlg;
static HFONT hEditorFont = NULL;
static LOGFONT lfEditor;
static COLORREF crEditorTextColor = 0;  //black
static int tabWidthChars = 4;
static int tabWidthTwips = -1;  //not yet set
static const char *szFontKey = "Software\\AFH Systems\\FilterMeister\\Editor\\Font";

static HWND hTT;      //tooltip control handle

int FM_is_busy = FALSE;

/* Forward prototypes */

INT_PTR WINAPI EditorDialogProc(HWND hDlg, UINT wMsg, WPARAM wParam, LPARAM lParam);

/*****************************************************************************/

static BOOLEAN SetEditorFont(void) {
    HFONT hNewFont;

#if 0
    //debugging...
    Info("Editor face name = %s\n"
         "Height = %d\n"
         "Width = %d\n"
         "Weight = %d\n"
         "Italic = %d\n"
         "Underline = %d\n"
         "StrikeOut = %d\n"
         "Character set = %d\n"
         "Out precision = %d\n"
         "Clip precision = %d\n"
         "Quality = %d\n"
         "Pitch and Family = 0x%.2X\n"
         , lfEditor.lfFaceName
         , lfEditor.lfHeight
         , lfEditor.lfWidth
         , lfEditor.lfWeight
         , lfEditor.lfItalic
         , lfEditor.lfUnderline
         , lfEditor.lfStrikeOut
         , lfEditor.lfCharSet
         , lfEditor.lfOutPrecision
         , lfEditor.lfClipPrecision
         , lfEditor.lfQuality
         , lfEditor.lfPitchAndFamily
         );
#endif
    hNewFont = CreateFontIndirect(&lfEditor);
    if (!hNewFont) {
#ifdef _DEBUG
        ErrorOk("CreateFontIndirect(&lfEditor) failed!");
#endif //_DEBUG
        return FALSE;
    }
    //Set font for the compiler output window...
    SendMessage(hEdit1, WM_SETFONT,
                (WPARAM) hNewFont,
                FALSE /*redraw flag*/);

    //Set font for the source editor window...
    SendMessage(hEdit2, WM_SETFONT,
                (WPARAM) hNewFont,
                FALSE /*redraw flag*/);

    if (hEditorFont) {
        //delete previous editor font
        DeleteObject(hEditorFont);
    }
    hEditorFont = hNewFont;

    {   //For good measure, also set default character formatting
        //for Rich Edit control...
        //NECESSARY TO CHANGE CHARSET!!!! (SetFont doesn't do it!)
        LRESULT status;
        CHARFORMAT cf;
        cf.cbSize = sizeof(cf);
        cf.dwMask = //CFM_BOLD | 
                    //CFM_ITALIC | 
                    CFM_UNDERLINE |
                    CFM_STRIKEOUT | 
                    //CFM_PROTECTED |
                    //CFM_LINK |
                    //CFM_SIZE |
                    CFM_COLOR | 
                    CFM_FACE | 
                    //CFM_OFFSET |
                    CFM_CHARSET |
                    0;
        cf.dwEffects = 0;
        if (lfEditor.lfWeight >= FW_BOLD) cf.dwEffects |= CFE_BOLD; //not enabled
        if (lfEditor.lfItalic) cf.dwEffects |= CFE_ITALIC; //not enabled
        if (lfEditor.lfUnderline) cf.dwEffects |= CFE_UNDERLINE;
        if (lfEditor.lfStrikeOut) cf.dwEffects |= CFE_STRIKEOUT;
        cf.yHeight = lfEditor.lfHeight; //not enabled; need to convert units?  Take ABS?
        cf.yOffset = 0; //not enabled
        cf.crTextColor = crEditorTextColor;
        cf.bCharSet = lfEditor.lfCharSet;
        cf.bPitchAndFamily = lfEditor.lfPitchAndFamily;
        strcpy(cf.szFaceName, lfEditor.lfFaceName);
        //set default char format...
        status = SendMessage(hEdit2, EM_SETCHARFORMAT, SCF_DEFAULT, (LPARAM)&cf);
#if 0
        //debugging...
        Info("SETCHARFORMAT(SCF_DEFAULT) status = %8.8Ix", status);
#endif
        //also apply to ALL existing chars...
        status = SendMessage(hEdit2, EM_SETCHARFORMAT, SCF_ALL, (LPARAM)&cf);
#if 0
        //debugging...
        Info("SETCHARFORMAT(SCF_ALL) status = %8.8Ix", status);
#endif
#if 0
        //debugging...
        //Get default char formatting...
        status = SendMessage(hEdit2, EM_GETCHARFORMAT, 0, (LPARAM)&cf);
        Info("GETCHARFORMAT\n"
            "status = %8.8Ix\n"
            "cbSize = %d\n"
            "dwMask = %8.8x\n"
            "dwEffects = %8.8x\n"
            "yHeight = %d\n"
            "yOffset = %d\n"
            "crTextColor = %8.8x\n"
            "bCharSet = %d\n"
            "bPitchAndFamily = 0x%.2X\n"
            "szFaceName = '%s'"
            ,
            status, 
            cf. cbSize,
            cf. dwMask,
            cf. dwEffects,
            cf. yHeight,
            cf. yOffset,
            cf. crTextColor,
            cf. bCharSet,
            cf. bPitchAndFamily,
            cf. szFaceName);
#endif
    }

    // Set tab stops for Code windows
    {   DWORD dwTabs = {16}; //16 dlu = 4 chars

        Edit_SetTabStops(hEdit1, 1, &dwTabs);

#ifdef USE_EDIT_CONTROL
        Edit_SetTabStops(hEdit2, 1, &dwTabs);
#else

        {   //Set default paragraph formatting for Rich Edit control
            LRESULT status;
            int i;
            HDC hdc = GetDC(hEdit2);   // handle of editor device context
            HFONT hPrevFont;
            TEXTMETRIC tm;
            int cxDpi;
            int newTabWidthTwips;
            CHARRANGE saveSel;
            static const CHARRANGE allSel = {0, -1};
            PARAFORMAT pf;
            pf.cbSize = sizeof(pf);

            //Compute tab width in Twips...
            SaveDC(hdc);
            SetGraphicsMode(hdc, GM_ADVANCED);
            ModifyWorldTransform(hdc, NULL, MWT_IDENTITY);
            SetViewportOrgEx(hdc, 0, 0, NULL);
            SetWindowOrgEx(hdc, 0, 0, NULL);

            cxDpi = GetDeviceCaps(hdc, LOGPIXELSX);
            if (cxDpi == 0) cxDpi = 96; //default to avoid zero div

            hPrevFont = SelectObject(hdc, hEditorFont);
            status = GetTextMetrics(hdc, &tm);
            SelectObject(hdc, hPrevFont);

            //twips = chars x pixels/char x twips/inch / pixels/inch
            newTabWidthTwips = tabWidthChars * tm.tmAveCharWidth * 1440 / cxDpi;

#if 0
            //debugging...
            Info("TEXTMETRICS\n"
                "status = %8.8Ix\n"
                "hdc = %8.8x\n"
                "cxDpi = %d\n"
                "tabWidthTwips = %d\n"
                "tmHeight = %d\n"
                "tmAscent = %d\n"
                "tmDescent = %d\n"
                "tmInternalLeading = %d\n"
                "tmExternalLeading = %d\n"
                "tmAveCharWidth = %d\n"
                "tmMaxCharWidth = %d\n"
                "tmWeight = %d\n"
                "tmOverhang = %d\n"
                , status
                , hdc
                , cxDpi
                , tabWidthTwips
                , tm.tmHeight           //pixels
                , tm.tmAscent           //pixels
                , tm.tmDescent          //pixels
                , tm.tmInternalLeading  //pixels
                , tm.tmExternalLeading  //pixels
                , tm.tmAveCharWidth     //pixels
                , tm.tmMaxCharWidth     //pixels
                , tm.tmWeight
                , tm.tmOverhang         //pixels
                );
#endif

            RestoreDC(hdc, -1);

            if (newTabWidthTwips != tabWidthTwips)
            {
                // TO AVOID UNNECESSARY FLASHING, DO THIS ONLY IF
                // tabWidthTwips has changed!!!

                tabWidthTwips = newTabWidthTwips;
#if 0
                //debugging...
                status = SendMessage(hEdit2, EM_GETPARAFORMAT, 0, (LPARAM)&pf);
                Info("GETPARAFORMAT\n"
                    "status = %8.8Ix\n"
                    "cbSize = %d\n"
                    "dwMask = %8.8x\n"
                    "wNumbering = %4.4x\n"
                    "wReserved = %4.4x\n"
                    "dxStartIndent = %d\n"
                    "dxRightIndent = %d\n"
                    "dxOffset = %d\n"
                    "wAlignment = %d\n"
                    "cTabCount = %d\n"
                    "Tab[0] = %d\n"
                    "Tab[1] = %d\n"
                    "Tab[2] = %d\n"
                    "Tab[3] = %d\n"
                    "Tab[4] = %d\n"
                    "Tab[5] = %d\n"
                    "Tab[6] = %d\n"
                    "Tab[7] = %d\n"
                    "Tab[8] = %d\n"
                    "Tab[9] = %d\n"
                    "Tab[10] = %d\n"
                    "Tab[11] = %d\n"
                    "Tab[12] = %d\n"
                    "Tab[13] = %d\n"
                    "Tab[14] = %d\n"
                    "Tab[15] = %d\n"
                    "Tab[16] = %d\n"
                    "Tab[17] = %d\n"
                    "Tab[18] = %d\n"
                    "Tab[19] = %d\n"
                    "Tab[20] = %d\n"
                    "...\n"
                    "Tab[31] = %d\n"
                    ,
                    status, 
                    pf. cbSize,
                    pf. dwMask,
                    pf. wNumbering,
                    pf. wReserved,
                    pf. dxStartIndent,
                    pf. dxRightIndent,
                    pf. dxOffset,
                    pf. wAlignment,
                    pf. cTabCount,
                    pf.rgxTabs[0],
                    pf.rgxTabs[1],
                    pf.rgxTabs[2],
                    pf.rgxTabs[3],
                    pf.rgxTabs[4],
                    pf.rgxTabs[5],
                    pf.rgxTabs[6],
                    pf.rgxTabs[7],
                    pf.rgxTabs[8],
                    pf.rgxTabs[9],
                    pf.rgxTabs[10],
                    pf.rgxTabs[11],
                    pf.rgxTabs[12],
                    pf.rgxTabs[13],
                    pf.rgxTabs[14],
                    pf.rgxTabs[15],
                    pf.rgxTabs[16],
                    pf.rgxTabs[17],
                    pf.rgxTabs[18],
                    pf.rgxTabs[19],
                    pf.rgxTabs[20],
                    pf.rgxTabs[31]);
#endif
                pf.dwMask = PFM_TABSTOPS;
                pf.cTabCount = MAX_TAB_STOPS;
                for (i = 0; i < MAX_TAB_STOPS; ++i) {
                    pf.rgxTabs[i] = tabWidthTwips*(i+1);
                }

                //save current selection
                SendMessage(hEdit2, EM_EXGETSEL, 0, (LPARAM)&saveSel);
                //select all source text
                SendMessage(hEdit2, EM_EXSETSEL, 0, (LPARAM)&allSel);

                //Set paragraph formatting for all text...
                status = SendMessage(hEdit2, EM_SETPARAFORMAT, 0, (LPARAM)&pf);

                //restore saved selection
                SendMessage(hEdit2, EM_EXSETSEL, 0, (LPARAM)&saveSel);
#if 0
                //debugging...
                Info("SETPARAFORMAT status = %8.8Ix", status);
                status = SendMessage(hEdit2, EM_GETPARAFORMAT, 0, (LPARAM)&pf);
                Info("GETPARAFORMAT\n"
                    "status = %8.8Ix\n"
                    "cbSize = %d\n"
                    "dwMask = %8.8x\n"
                    "wNumbering = %4.4x\n"
                    "wReserved = %4.4x\n"
                    "dxStartIndent = %d\n"
                    "dxRightIndent = %d\n"
                    "dxOffset = %d\n"
                    "wAlignment = %d\n"
                    "cTabCount = %d\n"
                    "Tab[0] = %d\n"
                    "Tab[1] = %d\n"
                    "Tab[2] = %d\n"
                    "Tab[3] = %d\n"
                    "Tab[4] = %d\n"
                    "Tab[5] = %d\n"
                    "Tab[6] = %d\n"
                    "Tab[7] = %d\n"
                    "Tab[8] = %d\n"
                    "Tab[9] = %d\n"
                    "Tab[10] = %d\n"
                    "Tab[11] = %d\n"
                    "Tab[12] = %d\n"
                    "Tab[13] = %d\n"
                    "Tab[14] = %d\n"
                    "Tab[15] = %d\n"
                    "Tab[16] = %d\n"
                    "Tab[17] = %d\n"
                    "Tab[18] = %d\n"
                    "Tab[19] = %d\n"
                    "Tab[20] = %d\n"
                    "...\n"
                    "Tab[31] = %d\n"
                    ,
                    status, 
                    pf. cbSize,
                    pf. dwMask,
                    pf. wNumbering,
                    pf. wReserved,
                    pf. dxStartIndent,
                    pf. dxRightIndent,
                    pf. dxOffset,
                    pf. wAlignment,
                    pf. cTabCount,
                    pf.rgxTabs[0],
                    pf.rgxTabs[1],
                    pf.rgxTabs[2],
                    pf.rgxTabs[3],
                    pf.rgxTabs[4],
                    pf.rgxTabs[5],
                    pf.rgxTabs[6],
                    pf.rgxTabs[7],
                    pf.rgxTabs[8],
                    pf.rgxTabs[9],
                    pf.rgxTabs[10],
                    pf.rgxTabs[11],
                    pf.rgxTabs[12],
                    pf.rgxTabs[13],
                    pf.rgxTabs[14],
                    pf.rgxTabs[15],
                    pf.rgxTabs[16],
                    pf.rgxTabs[17],
                    pf.rgxTabs[18],
                    pf.rgxTabs[19],
                    pf.rgxTabs[20],
                    pf.rgxTabs[31]);
#endif
            } //if tabWidthTwips changed
        }//rich edit
#endif //USE_EDIT_CONTROL
    }//set tab stops

    return TRUE; //success!
} //SetEditorFont

/*****************************************************************************/

static BOOLEAN ChangeEditorFont(HWND hDlg) {
    {   //Choose new font...
        CHOOSEFONT cf;

        cf.lStructSize = sizeof(cf);
        cf.hwndOwner = hDlg;
        cf.hDC = NULL;
        cf.lpLogFont = &lfEditor;
        cf.iPointSize = 0;  //output param
        cf.Flags = CF_INITTOLOGFONTSTRUCT | CF_SCREENFONTS | CF_EFFECTS;
        cf.rgbColors = crEditorTextColor;
        cf.lCustData = 0;
        cf.lpfnHook = NULL;
        cf.lpTemplateName = NULL;
        cf.hInstance = NULL;
        cf.lpszStyle = NULL;
        cf.nFontType = 0;   //output param
        cf.nSizeMin = 0;
        cf.nSizeMax = 0;
        if (!ChooseFont(&cf)) {
            //User canceled or closed the dialog...
            return FALSE;
        }
        crEditorTextColor = cf.rgbColors;
    }//choose font

    //Set the font and tabs...
    if (!SetEditorFont()) {
        //failed...
        return FALSE;
    }

    //Invalidate Editor client rect to force reformat???
    //Doesn't seem to be necessary!


    // Save font and tabs in registry...
    {
        RSS(RCK(HKEY_CURRENT_USER, szFontKey), "FaceName", lfEditor.lfFaceName);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "Height", -lfEditor.lfHeight);   // N.B. < 0 for char height
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "Width", lfEditor.lfWidth);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "Weight", lfEditor.lfWeight);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "Italic", lfEditor.lfItalic);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "Underline", lfEditor.lfUnderline);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "StrikeOut", lfEditor.lfStrikeOut);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "CharSet", lfEditor.lfCharSet);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "OutPrecision", lfEditor.lfOutPrecision);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "ClipPrecision", lfEditor.lfClipPrecision);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "Quality", lfEditor.lfQuality);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "PitchAndFamily", lfEditor.lfPitchAndFamily);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "TextColor", crEditorTextColor);
        RSD(RCK(HKEY_CURRENT_USER, szFontKey), "TabWidthChars", tabWidthChars);
    }

    return TRUE;    //Success!
} //ChangeEditorFont

/*****************************************************************************/


void EditorThread(void *arg)
{
    // Create a modeless dialog box
    MSG msg;
    HACCEL hAccel;

#ifndef USE_EDIT_CONTROL
    {
        HINSTANCE hREdll =
    #if 1
        LoadLibrary("RICHED32.DLL");
    #elif 0
        LoadLibrary("RICHED20.DLL");    //NFG - use with "RichEdit20A" or "RichEdit20W" ???
    #elif 0
        LoadLibrary("RICHED.DLL");      //NFG
    #endif
        if (!hREdll) {
            ErrorOk("Can't load Rich Edit DLL.");
        }
    }
#endif

    hEditorDlg = CreateDialogParam(
		               hDllInstance,
                       (LPSTR)"EDITOR",
    #if 0
                       NULL,        //no owner
    #else
                       (HWND)MyHDlg,
    #endif
                       (DLGPROC)EditorDialogProc,
                       (LPARAM)0);

    // Set Window Theme for the dialog and/or all its controls???

    //Show the dialog
    ShowWindow(hEditorDlg, SW_SHOW);

    //Get Accelerators for Edit dialog
    hAccel = LoadAccelerators(hDllInstance, MAKEINTRESOURCE(IDR_ACCELERATOR_EDIT));

    //Begin a message pump...
    while (GetMessage(&msg, NULL, 0, 0)) {
	    // Need to translate accelerators *before* calling
        // IsDialogMessage.  Otherwise, the dialog eats
        // the WM_KEYDOWN etc. messages.
        if (TranslateAccelerator(hEditorDlg, hAccel, &msg)) {
            continue;
        }
        if (IsDialogMessage(hEditorDlg, &msg)) {
            continue;
        }
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }//while

    //delete editor font if created...
    if (hEditorFont) {
#ifdef _DEBUG
        if (!DeleteObject(hEditorFont)) {
            ErrorOk("DeleteObject(hEditorFont) failed!");
        }
#else
        DeleteObject(hEditorFont);
#endif
        hEditorFont = NULL;
    }

    //need to destroy the accelerator table here?
    //--probably not necessary, but appears to be harmless...
    if (!DestroyAcceleratorTable(hAccel)) {
        ErrorOk("Failed to destroy accelerator table.");
    }

    //need to destroy the modeless dialog box here?
    if (hEditorDlg) {
        DestroyWindow(hEditorDlg);
        hEditorDlg = 0;
    }

    //need to free the RichEdit library???
    ///// FreeLibrary(hREdll);   ????


    _endthread();
} //EditorThread


/*****************************************************************************/

void StartEditor(void)
{
    if (!hEditorDlg) {
        _beginthread(EditorThread, 0, NULL);
    }
}

/*****************************************************************************/
// Edit control subclass procedure

static WNDPROC wpOrigEditProc;

static LRESULT WINAPI EditSubclassProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{ 
    //so we can handle TAB chars...
    if (wMsg == WM_GETDLGCODE)
        return DLGC_WANTALLKEYS;

    return CallWindowProc(wpOrigEditProc, hwnd, wMsg, wParam, lParam);

} /*EditSubclassProc*/
    
/*****************************************************************************/
// Register a tooltip for an Editor dialog control

static void RegisterEditorTooltip(HWND hDlg, int ctl, LPSTR text)
{   // Set tooltips for Editor Buttons...
    TOOLINFO ti;

    //register a tooltip for the designated control
    ti.cbSize = sizeof(TOOLINFO);
    ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
    ti.hwnd = hDlg;
    ti.uId = (UINT_PTR) GetDlgItem(hDlg, ctl);
    ti.rect.left = ti.rect.top = ti.rect.right = ti.rect.bottom = 0;
    ti.hinst = hDllInstance;
    ti.lpszText = text;
    SendMessage(hTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
} /*RegisterEditorTooltip*/

/*****************************************************************************/
// Dialog Procedure for the FM Editor

INT_PTR WINAPI EditorDialogProc(HWND hDlg, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
    int idd;
    int iCmd;
    static HWND hwndEdit;

    switch  (wMsg) {

        case  WM_INITDIALOG:
            FM_is_busy = FALSE;
#ifdef FULL
            hEdit1 = GetDlgItem(hDlg, IDC_EDIT1);
#endif
            hEdit2 = GetDlgItem(hDlg, IDC_EDIT2);

            {   // Restore previous dialog position unless first time...
                if (gParams->gEditorRect.bottom == 0) {
                    // First time, position below the FM UI dialog...
                    RECT    DlgRect;
                    RECT    rcDialog;
                    RECT    rcClient;
                    RECT    r;
                    /* Get FM UI dialog coordinates */
                    GetWindowRect(MyHDlg, &DlgRect);
                    /* Convert dialog units to pixels */
                    r.left = 0;
                    r.top = 0;

                    //Changed by Harald Heim, March 9, 2003
                    r.right = 485; //388;  //initial size of Editor dialog in DBU's
                    r.bottom = 207; //173;
                    
                    MapDialogRect(hDlg, &r);
                    // Need to add SM_CYCAPTION, SM_CXDLGFRAME, etc.
                    // Add the difference between dialog window size and dialog client
                    // size, so it automatically adjusts for caption, border, etc.
                    GetWindowRect(hDlg, &rcDialog);
                    GetClientRect(hDlg, &rcClient);
                    r.right += (rcDialog.right-rcDialog.left) - (rcClient.right-rcClient.left);
                    r.bottom += (rcDialog.bottom-rcDialog.top) - (rcClient.bottom-rcClient.top);
                    SetWindowPos(hDlg, NULL,
                                 DlgRect.left,
                                 DlgRect.bottom + 5,//10, //Changed by Harald Heim, March 21, 2003
                                 r.right,       //width
                                 r.bottom,      //height
                                 SWP_NOZORDER | SWP_NOACTIVATE);
                }
                else {
                    //On subsequent invocations, use coordinates that were
                    //saved at previous exit...
                    SetWindowPos(hDlg, NULL,
                                 gParams->gEditorRect.left,
                                 gParams->gEditorRect.top,
                                 gParams->gEditorRect.right - gParams->gEditorRect.left,
                                 gParams->gEditorRect.bottom - gParams->gEditorRect.top,
                                 SWP_NOZORDER | SWP_NOACTIVATE);
                }
            }

            // Create the Tool Tip control first!
            InitCommonControls();

            hTT = CreateWindowEx(  
                0,         // extended styles
                TOOLTIPS_CLASS,   // predefined class 
                NULL,       // text 
                WS_VISIBLE | WS_CHILD| WS_POPUP | TTS_ALWAYSTIP,  // styles 
                CW_USEDEFAULT,         // starting x position 
                CW_USEDEFAULT,         // starting y position 
                CW_USEDEFAULT,        // button width 
                CW_USEDEFAULT,        // button height 
                hDlg,       // parent window 
                NULL,       // TT control ID
                hDllInstance,
                NULL);      // pointer not needed 

            // Per MSDN, we must explicitly define a ToolTip
            // control as topmost...
#if 0
            // but it screws up z-pos of common dialogs such as Load/Save!!!
            SetWindowPos(hTT, HWND_TOPMOST, 0, 0, 0, 0,
                         SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
#endif


            // Set fonts and tabs for Code windows

            //Set default font attributes and tab stops...
            /// Actually only need to do when first loaded????
            GetObject(GetStockObject(ANSI_FIXED_FONT), sizeof(lfEditor), &lfEditor);
            //crEditorTextColor = 0; //black
            tabWidthTwips = -1; //force tab updates
            //tabWidthChars = 4;

            //Now override attributes with registry entries, if any...
            /// Actually only need to do when first loaded????
            {   
                DWORD dwValue;
                char szFaceName[LF_FACESIZE];

                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "FaceName", szFaceName, sizeof(szFaceName)) == ERROR_SUCCESS)
                {
                    strncpy(lfEditor.lfFaceName, szFaceName, sizeof(lfEditor.lfFaceName));
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "Height", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfHeight = -(int)dwValue;   // N.B. < 0 for char height
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "Width", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfWidth = dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "Weight", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfWeight = dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "Italic", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfItalic = (BYTE)dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "Underline", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfUnderline = (BYTE)dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "StrikeOut", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfStrikeOut = (BYTE)dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "CharSet", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfCharSet = (BYTE)dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "OutPrecision", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfOutPrecision = (BYTE)dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "ClipPrecision", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfClipPrecision = (BYTE)dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "Quality", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfQuality = (BYTE)dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "PitchAndFamily", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    lfEditor.lfPitchAndFamily = (BYTE)dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "TextColor", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    crEditorTextColor = dwValue;
                }
                if (RQV(ROK(HKEY_CURRENT_USER, szFontKey),
                    "TabWidthChars", (LPBYTE)&dwValue, sizeof(dwValue)) == ERROR_SUCCESS)
                {
                    //make sure it's reasonable!
                    if (dwValue > 0 && dwValue < 100) tabWidthChars = dwValue;
                }
            }//registry overrides

            //Set the fonts and tabs...
            SetEditorFont();



#ifdef USE_EDIT_CONTROL
    #if 1
            // Try to increase size of Edit2 beyond the default of 30,000 chars...
            Edit_LimitText(hEdit2, 0);
    #else
            // Try to increase size of Edit2 beyond the default of 30,000 chars...
            Edit_LimitText(hEdit2, 65000);
    #endif

    #if 1
            // Try to force initial allocation of text buffer in edit control
            // to be large enough to handle max source code size...
            {
                char *lpsz;
                lpsz = _alloca(MAX_SOURCE_CODE_SIZE + 1);
                if (lpsz) {
                    memset(lpsz, '*', MAX_SOURCE_CODE_SIZE);
                    lpsz[MAX_SOURCE_CODE_SIZE] = 0; //terminating NUL
                    Edit_SetText(hEdit2, lpsz);
                }
            }
    #endif
#else
            //Default limit for a Rich Edit control is 32K, so we need
            //to increase it...
            SendMessage(hEdit2, EM_EXLIMITTEXT, 0, MAX_SOURCE_CODE_SIZE);
#endif  //USE_EDIT_CONTROL

            // Restore saved source code to source code edit window...
            SetWindowText(hEdit2, gParams->szSrcCode);
            // And set cursor to top of buffer or to saved position????

#ifndef USE_EDIT_CONTROL
            //Enable EN_CHANGE notification messages from the Rich Edit
            //control (or just use EM_GETMODIFY????).
            SendMessage(hEdit2, EM_SETEVENTMASK, 0,
                        SendMessage(hEdit2, EM_GETEVENTMASK, 0, 0) | ENM_CHANGE);
#endif

            if (!(gParams->flags & AFH_EDITMODE)) {
            }
#ifdef FULL
            else if (gParams->flags & AFH_ADVANCED) {
              SetDlgItemText(hDlg, IDC_ADVANCED, "&<<<");
              SetDlgItemText(hDlg, IDC_GO, "Pre&view");
            }
#endif //FULL
            else {
#ifdef FULL
              SetDlgItemText(hDlg, IDC_ADVANCED, "&>>>");
#endif
              SetDlgItemText(hDlg, IDC_GO, "&Compile");
            }

			SetDlgItemText(hDlg, IDC_X64GEN, X64GEN?"&X64":"&X32");

            //Enable/disable Edit mode buttons...
            EnableWindow(GetDlgItem(hDlg, IDC_LOAD), (gParams->flags & AFH_EDITMODE) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_SAVE), (gParams->flags & AFH_EDITMODE) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_FONT), (gParams->flags & AFH_EDITMODE) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_X64GEN), (gParams->flags & AFH_EDITMODE) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_GO), (gParams->flags & AFH_EDITMODE) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_MAKE), (gParams->flags & AFH_EDITMODE) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_HELP1), (gParams->flags & AFH_EDITMODE) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_HELP2), (gParams->flags & AFH_EDITMODE) != 0);

#ifdef FULL
            EnableWindow(GetDlgItem(hDlg, IDC_ADVANCED), (gParams->flags & AFH_EDITMODE) != 0);
            //Show/Hide Advanced mode buttons..
#if 1 //IDC_PREPROCESS is disabled for now
            EnableWindow(GetDlgItem(hDlg, IDC_PREPROCESS), 0);
#else
            EnableWindow(GetDlgItem(hDlg, IDC_PREPROCESS), (gParams->flags & AFH_ADVANCED) != 0);
#endif
            EnableWindow(GetDlgItem(hDlg, IDC_PARSE), (gParams->flags & AFH_ADVANCED) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_COMPILE), (gParams->flags & AFH_ADVANCED) != 0);
            EnableWindow(GetDlgItem(hDlg, IDC_INIT), (gParams->flags & AFH_ADVANCED) != 0);
            ShowWindow(GetDlgItem(hDlg, IDC_PREPROCESS), gParams->flags & AFH_ADVANCED?SW_SHOW:SW_HIDE);
            ShowWindow(GetDlgItem(hDlg, IDC_PARSE), gParams->flags & AFH_ADVANCED?SW_SHOW:SW_HIDE);
            ShowWindow(GetDlgItem(hDlg, IDC_COMPILE), gParams->flags & AFH_ADVANCED?SW_SHOW:SW_HIDE);
            ShowWindow(GetDlgItem(hDlg, IDC_INIT), gParams->flags & AFH_ADVANCED?SW_SHOW:SW_HIDE);
#endif //FULL

            SetDlgItemInt(hDlg, 100, gParams->bandWidth, TRUE /*signed*/); 

            // Enable/disable "Use MMX", "Use CMOV" and "User_SSE" checkboxes...
            EnableWindow(GetDlgItem(hDlg, IDC_MMX), X86_MMX_available());
            EnableWindow(GetDlgItem(hDlg, IDC_USE_CMOV), X86_CMOV_available());
 ///           EnableWindow(GetDlgItem(hDlg, IDC_USE_SSE), X86_SSE_available());	//NYI

            //Enable/disable Bandwidth edit box...
            EnableWindow(GetDlgItem(hDlg, 100), gFmcp->isTileable);

            // Initialize "system" checkboxes...
            SendMessage(GetDlgItem(hDlg, IDC_DRAGOUT),
                        BM_SETCHECK, (gParams->flags & AFH_DRAG_FILTERED_OUTPUT) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_TRACE),
                        BM_SETCHECK, (gParams->flags & AFH_TRACE) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_MMX),
                        BM_SETCHECK, (gParams->flags & AFH_USE_MMX) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_ADVSTATE),
                        BM_SETCHECK, (gParams->flags & AFH_USE_ADVANCESTATE) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_HOSTPAD),
                        BM_SETCHECK, (gParams->flags & AFH_USE_HOST_PADDING) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_OPTZOOM),
                        BM_SETCHECK, (gParams->flags & AFH_INIT_OPTIMAL_ZOOM) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_BIGGULP),
                        BM_SETCHECK, (gParams->flags & AFH_PROXY_BIG_GULP) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_FLUSHA),
                        BM_SETCHECK, (gParams->flags & AFH_FLUSH_ADVANCESTATE) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_RGB),
                        BM_SETCHECK, (gParams->flags & AFH_SUPPORT_RGB) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_GRAY),
                        BM_SETCHECK, (gParams->flags & AFH_SUPPORT_GRAY) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_CMYK),
                        BM_SETCHECK, (gParams->flags & AFH_SUPPORT_CMYK) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_LAB),
                        BM_SETCHECK, (gParams->flags & AFH_SUPPORT_LAB) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_INDEXED),
                        BM_SETCHECK, (gParams->flags & AFH_SUPPORT_INDEXED) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_BITMAP),
                        BM_SETCHECK, (gParams->flags & AFH_SUPPORT_BITMAP) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_DUOTONE),
                        BM_SETCHECK, (gParams->flags & AFH_SUPPORT_DUOTONE) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_MULTI),
                        BM_SETCHECK, (gParams->flags & AFH_SUPPORT_MULTI) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_ZDRAG),
                        BM_SETCHECK, (gParams->flags & AFH_ZERO_OUTPUT_DRAG) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_DISASSEMB),
                        BM_SETCHECK, (gParams->flags & AFH_DISASSEMBLE) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_DUMPSDL),
                        BM_SETCHECK, (gParams->flags & AFH_DUMP_SDL) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_PSEUDOCODE),
                        BM_SETCHECK, (gParams->flags & AFH_PSEUDOCODE) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_USE_CMOV),
                        BM_SETCHECK, (gParams->flags & AFH_USE_CMOV) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_USE_SSE),
                        BM_SETCHECK, (gParams->flags & AFH_USE_SSE) != 0, 0);
            SendMessage(GetDlgItem(hDlg, IDC_USE_INLINES),
                        BM_SETCHECK, (gParams->flags & AFH_USE_INLINES) != 0, 0);

			// Set tooltips for Editor Buttons... (Some could vary..)
			// Open...
			RegisterEditorTooltip(hDlg, IDC_LOAD, "Load or import filter code");
			// Save...
			RegisterEditorTooltip(hDlg, IDC_SAVE, "Save filter source code");
			// Font...
			RegisterEditorTooltip(hDlg, IDC_FONT, "Change source editor font");
			// AutoLoad (could vary)
			RegisterEditorTooltip(hDlg, IDC_AUTOLOAD,
				"Reload and compile when source file is modified by an external editor");
			// X32/X64 (could vary)
			RegisterEditorTooltip(hDlg, IDC_X64GEN, "Click to change target architecture");
			// Preprocess
			RegisterEditorTooltip(hDlg, IDC_PREPROCESS, "Preprocess macros and includes");
			// Parse
			RegisterEditorTooltip(hDlg, IDC_PARSE, "Parse the source code");
			// CodeGen
			RegisterEditorTooltip(hDlg, IDC_COMPILE, "Generate machine code");
			// Initialize
			RegisterEditorTooltip(hDlg, IDC_INIT, "Initialize the filter");
			// Compile/Preview (could vary)
			RegisterEditorTooltip(hDlg, IDC_GO, "Compile and run filter source code");
			// Make...
			RegisterEditorTooltip(hDlg, IDC_MAKE, "Make a standalone filter");
			// About...
			RegisterEditorTooltip(hDlg, IDC_HELP2, "Display filter information and copyright");
			// Advanced >>>... (could vary)
			RegisterEditorTooltip(hDlg, IDC_ADVANCED, "Toggle advanced mode display");
			// Help...
			RegisterEditorTooltip(hDlg, IDC_HELP1, "Display help in a web browser");

            // Subclass the IDC_EDIT2 edit control...
            // Retrieve the handle of the edit control. 
            hwndEdit = hEdit2;  
            wpOrigEditProc = (WNDPROC) SetWindowLongPtr(hwndEdit, 
                                        GWLP_WNDPROC, (LONG_PTR) EditSubclassProc);

            // Set Focus to Source Code Edit Text Item
            SetFocus(hEdit2);

            return  FALSE; // FALSE says we set the focus to a control.

        

        case WM_ACTIVATE:
            //Forward Message
            //if ( LOWORD(wParam) == WA_CLICKACTIVE && (MyHDlg != (HWND) lParam) && (hDlg != (HWND) lParam)){ 
            if ( LOWORD(wParam) > 0 && (MyHDlg != (HWND) lParam) && (hDlg != (HWND) lParam)){ 
                PostMessage(MyHDlg, wMsg, wParam, lParam);
            }
            return 0;


        case WM_MOUSEMOVE:
        case WM_LBUTTONDOWN:
        case WM_LBUTTONUP:
        case WM_RBUTTONDOWN:
        case WM_RBUTTONUP:
        case WM_MBUTTONDOWN:
        case WM_MBUTTONUP:
            {
                // relay this mouse message to the tooltip control...
                MSG msg; 

                msg.lParam = lParam; 
                msg.wParam = wParam; 
                msg.message = wMsg; 
                msg.hwnd = hDlg; 
                SendMessage(hTT, TTM_RELAYEVENT, 0, (LPARAM) &msg); 
            }
            return FALSE;
            break;

        case WM_ERASEBKGND:
            {   // Do a gradient fill a la Prosise BitmapDemo.cpp...
                HBRUSH hBrush[64];
                int i;
                int nWidth;
                int nHeight;
                RECT rect;
                // MUST use the DC passed in wParam, not the DC of hDlg --
                // The clipping/update regions are different!!!
                HDC hDC = (HDC) wParam; // handle of device context

                //COLORREF color1 = RGB(255,0,0);
                //COLORREF color2 = RGB(0,255,0);
    
                //Added by Harald Heim, March 9, 2003
                COLORREF color1 =  GetSysColor(COLOR_BTNFACE);//RGB(0xD4,0xD0,0xC8);  //yellow grey
                COLORREF color2 =  GetSysColor(COLOR_BTNFACE);//RGB(0xD4,0xD0,0xC8);   //yellow grey
                //COLORREF color1 = RGB(0x5F,0x9E,0xA0);   //X11.cadetblue
                //COLORREF color2 = RGB(0x5F,0x9E,0xA0);   //X11.cadetblue

                int r1 = GetRValue(color1);
                int g1 = GetGValue(color1);
                int b1 = GetBValue(color1);
                int r2 = GetRValue(color2);
                int g2 = GetGValue(color2);
                int b2 = GetBValue(color2);

                GetClientRect(hDlg, &rect);
                nWidth = rect.right - rect.left;
                nHeight = rect.bottom - rect.top;

                if (color1 == color2) {
                    // optimize this if color1 == color2
                    hBrush[0] = CreateSolidBrush(color1);
                    FillRect(hDC, &rect, hBrush[0]);
                    DeleteObject(hBrush[0]);
                }
                else
                {
                    // create a 64-step linear gradient
                    for (i=0; i < 64; i++)
                    {
                        hBrush[i] = CreateSolidBrush(
                            RGB( (r1*(63-i)+r2*i)/63, (g1*(63-i)+g2*i)/63, (b1*(63-i)+b2*i)/63 )
                            );
                    }

                    if (gParams->gDialogGradientDirection == 0) {
                        // vertical gradient
                        for (i=0; i < 64; i++) {
                            SetRect(&rect, 0, i*nHeight/64, nWidth, (i+1)*nHeight/64);
                            FillRect(hDC, &rect, hBrush[i]);
                        }
                    }
                    else {
                        // horizontal gradient
                        for (i=0; i < 64; i++) {
                            SetRect(&rect, i*nWidth/64, 0, (i+1)*nWidth/64, nHeight);
                            FillRect(hDC, &rect, hBrush[i]);
                        }
                    }

                    for (i=0; i < 64; i++)
                        DeleteObject(hBrush[i]);
                } /* color1 != color2 */

            }
            return TRUE;   // No more background painting needed.

        case WM_CTLCOLORSTATIC:
        case WM_CTLCOLORBTN:
        case WM_CTLCOLORLISTBOX:
        case WM_CTLCOLORSCROLLBAR:
#if 0
            // Yes, this controls nonreadonly edit box colors;
            // Readonly edit boxes seem to use WM_CTLCOLORSTATIC instead.
        case WM_CTLCOLOREDIT:
#endif

            {
                HDC hdc = (HDC) wParam;   // handle of display context
                HWND hwnd = (HWND) lParam; // handle of static control
                int iCtl;

                iCtl = GetDlgCtrlID(hwnd);    // ID of control
 
                if (					//////////////////  ARRGHH!! MAINTENANCE NIGHTGMARE!!!
                       (iCtl >= 803 && iCtl <= 899) // Static text but NOT Group boxes
                    || (iCtl >= IDC_TRACE && iCtl <= IDC_USE_SSE)
                    || (iCtl >= IDC_DISASSEMB && iCtl <= IDC_USE_CMOV)
                    || (iCtl == IDC_USE_INLINES)
                    )
                {   // We want these controls to be transparent...
                    // Note that these controls are NOT updated dynamically
                    // (which doesn't work right, since the backgrounds don't
                    // get erased properly).
					SetBkMode(hdc, GetSysColor(COLOR_BTNFACE));//OPAQUE); ////TRANSPARENT);
                    SetTextColor(hdc, GetSysColor(COLOR_BTNTEXT) ); //RGB(255,255,255));    //white text
                    //should really use SetDlgMsgResult...
                    return (INT_PTR) (HBRUSH) CreateSolidBrush(GetSysColor(COLOR_BTNFACE));//GetStockObject(HOLLOW_BRUSH); // create statically?
                }
                else if (iCtl >= 800 && iCtl <= 802) { //Advanced mode group boxes
                    SetBkColor(hdc, GetSysColor(COLOR_BTNFACE));//RGB(128,128,128));  //gray background
                    SetTextColor(hdc, GetSysColor(COLOR_BTNTEXT) ); //RGB(255,255,255));    //white text
                    //should really use SetDlgMsgResult...
                    return (INT_PTR) (HBRUSH) GetStockObject(HOLLOW_BRUSH);//GRAY_BRUSH);
                }
                else if (1)
                {
                    SetBkColor(hdc, RGB(255,255,255));  //white background
                    SetTextColor(hdc, RGB(0,0,100));    //dark blue text
                    //should really use SetDlgMsgResult...
                    return (INT_PTR) (HBRUSH) GetStockObject(WHITE_BRUSH); // create statically?
                }
                else return FALSE;  // Not processed.
            }
            break;
		
        case WM_CLOSE:
            PostMessage(MyHDlg, WM_COMMAND, IDC_EDIT, 0);
            return TRUE;   // don't call defproc

        case WM_DESTROY:
            // Remove the subclass from the edit control. 
            SetWindowLongPtr(hwndEdit, GWLP_WNDPROC, (LONG_PTR)wpOrigEditProc);
            FM_is_busy = FALSE;
            return FALSE;   // call defproc

        case WM_COMMAND:
            idd = COMMANDID (wParam);
            iCmd = COMMANDCMD(wParam, lParam);
            switch  (idd) {

            case IDC_SELECT_ALL:
                // Highlight all characters in the source edit window...
#ifdef USE_EDIT_CONTROL
                Edit_SetSel(hEdit2, 0, -1);
                // ...and scroll end of range into view if necessary!
                Edit_ScrollCaret(hEdit2);   //NFG with Rich Edit???
#else
                {
                    static const CHARRANGE ichCharRange = {0, -1};
                    SendMessage(hEdit2, EM_EXSETSEL, 0, (LPARAM)&ichCharRange);
                }
#endif //USE_EDIT_CONTROL
                return TRUE;

#ifdef FULL
            case IDC_FONT:
#ifndef USE_EDIT_CONTROL
                //Disable EN_CHANGE notification messages from the Rich Edit control.
                //MAY NOT WORK WITH LATER VERSIONS OF RICHEDIT!!!!
                SendMessage(hEdit2, EM_SETEVENTMASK, 0,
                            SendMessage(hEdit2, EM_GETEVENTMASK, 0, 0) & ~ENM_CHANGE);
#endif
                // Change editor font.
                ChangeEditorFont(hDlg);

#ifndef USE_EDIT_CONTROL
                //Re-enable EN_CHANGE notification messages from the Rich Edit control.
                SendMessage(hEdit2, EM_SETEVENTMASK, 0,
                            SendMessage(hEdit2, EM_GETEVENTMASK, 0, 0) | ENM_CHANGE);
#endif
                // Set Focus to Source Code Edit Text Item
                SetFocus (hEdit2);
                return TRUE;

            case IDC_MAKE:
                // Make a standalone filter.
                if (FM_is_busy) {
                    MessageBeep(MB_OK);
                    return TRUE;
                }
                FM_is_busy = TRUE;
                if (!FmMakeStandalone(hDlg))
                {
                    FM_is_busy = FALSE;
                    return FALSE;     //failed
                }
                FM_is_busy = FALSE;
                return TRUE;
#endif

            case IDOK:
            case IDCANCEL:
                break;          // fall through to EndDialog()

			case IDC_PREPROCESS:
            case IDC_PARSE:
            case IDC_COMPILE:
            case IDC_INIT:
            case IDC_GO:
            case IDC_HELP2:
            case IDC_HELP1: //Added by Harald Heim, March 22, 2003
            case IDC_LOAD:
            case IDC_SAVE:
			case IDC_X64GEN:
                if (FM_is_busy) {
                    MessageBeep(MB_OK);
                    return TRUE;
                }
                // Fall through to default case...

            default:
                if (hEditorDlg) {
                    // Forward this command to the filter dialog...
#if 0
                    SendMessage(MyHDlg, wMsg, wParam, lParam);
#else
                    PostMessage(MyHDlg, wMsg, wParam, lParam);
#endif
                }
                return FALSE;
            } //switch idd

            // Close the editor dialog...
            // Save current position and size of dialog
            GetWindowRect(hDlg, &gParams->gEditorRect);
            if (hEditorDlg) {
                DestroyWindow(hDlg);    //close modeless dialog
                hEditorDlg = 0;
            }
            return TRUE;    // do not call defProc

        default:
            return  FALSE;  // wMsg not handled
    } //switch wMsg

    return  TRUE;   // handled; do not call defProc

} /*EditorDialogProc*/

/*****************************************************************************/

#endif //!STANDALONE
