/*----------------------------------------------------------*/
/* AfhResUp.c                                               */
/*                                                          */
/* 12-May-1999  afh	1. Derived from AfhPED.c.               */
/* 14-AUG-2013	afh	1. Changes for WIN64.					*/
/* 10-DEC-2013  afh 1. Merge PE32+ changes from AfhPED 1.9. */
/* 04-MAR-2014  afh 1. "Bad fixup targets RVA 0x00000000"   */
/*					   warning message suppressed in 32-bit */
/*					   fixups. Seems to be harmless in x32  */
/*					   statically linked standalones.		*/
/*                                                          */
/*----------------------------------------------------------*/

// To do:
// - parse debug_misc, debug_fpo, etc.
// - parse forwarder chain in imports
// -  -b = dump bound import table
// -  -g = dump global ptr info
// -  -l = dump load config info
// -  -y = dump security info
// -  -c = dump copyright info
// - Is pFPO->ulOffStart an RVA; YES it is!!!
// - DOS stub begins after EXE relocations?
// - Dump EXE relocations (e.g., in XCOPY32.EXE)
// - Dump padding/unparsed bytes
// - adjust file offset of COFF symbols in PE header

// ifdef RAM_FILE...
// error return from ErrorHandler, ErrorOk???
// RESLEVEL is currently unneeded and should be deleted(?).

/////////////// HASH LIST IMPLEMENTATION NOTE //////////////////
//
// The hash list for name nodes is currently an exceedingly
// simple one -- there is only a single hash bucket (i.e.,
// the (trivial) hash function maps each name to the 0-indexed
// hash bucket)!  To wit:
// #define HASH(name) (0)
//
// This should be fixed before it ever becomes noticeably slow,
// both here and in AfhPED!!
//--------------------------------------------------------------

#include <windows.h>
//#include <commctrl.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <stddef.h>     /* offsetof */
#include <assert.h>
#include "AfhFM.h"

static FILE *fExe;
static FILE *fOut;
static LPCTSTR szFileName;      /* Initial, final file name */
static LPCTSTR szOutFileName;   /* Output file name (may be temporary) */
static LPCTSTR tempFileName = "temp.dll";

static size_t currentFilePos, lastFilePos;
static size_t SectTableOffset;
static size_t DosHeaderOffset;
static size_t PeHeaderOffset;
static size_t PeOptHeaderOffset;
static size_t ResRootOffset;
static int RESLEVEL = 0;		//no longer needed -- delete all references!!

static IMAGE_DOS_HEADER dosHeader;
static DWORD Signature;
static IMAGE_FILE_HEADER peHeader;
typedef union {	//union of PE32 and PE32+ optional headers
		IMAGE_OPTIONAL_HEADER32 ;	//anonymous member
		IMAGE_OPTIONAL_HEADER64 u64;
} IMAGE_COMMON_OPTIONAL_HEADER;
static IMAGE_COMMON_OPTIONAL_HEADER optHeader;
static IMAGE_COMMON_OPTIONAL_HEADER newOptHeader;
static IMAGE_SECTION_HEADER sectionTableEntry;
static IMAGE_DEBUG_DIRECTORY debugDirEntry;

/* Local copies of a few common header fields (preloaded) */
static UINT64 ImageBase = 0;
static DWORD NumberOfRvaAndSizes = 0;
/* Pointer to Data Directory */
static PIMAGE_DATA_DIRECTORY pDD;
/* Pointer to output Data Directory */
static PIMAGE_DATA_DIRECTORY oDD;
/* Pointer to Section Table */
static PIMAGE_SECTION_HEADER pST = NULL;
/* Pointer to output Section Table */
static PIMAGE_SECTION_HEADER oST = NULL;
/* Number of Section Table entries */
static int nST;
/* Presumed section number of .idata section */
static int idataSect;
/* Presumed section number of .rsrc section */
static int rsrcSect;
static BOOL bPE32PLUS = FALSE;
static BOOL bFixupsBeyondRsrc = FALSE;
static DWORD topRsrcRVA = 0;
static int outputRVAdelta = 0;
static DWORD topRsrcFileOffset = 0;
static int outputFileOffsetDelta = 0;
/* Table to record fixups to various sections */
typedef struct {
    int nFixupsToThisSection;
    int nFixupTargetsToThisSection;
} FIXUP_INFO;
static FIXUP_INFO *pFixups = NULL;

/* DOS Stub Program */
static size_t stubSize;
static void *pDosStub = NULL;

/* RAM file stuff */
BOOL bRamFile = FALSE;
char *pRamFile;


static void ErrorHandler(LPCTSTR eMsg) {
    ErrorOk(eMsg);
    //exit(1);
} //ErrorHandler

#if 0
static void SysErrorMessage(LPCTSTR prefix) {
    LPVOID lpMsgBuf;
    DWORD msgid = GetLastError();

    FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,	// pointer to  message source 
        msgid,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPTSTR) &lpMsgBuf,
        0,	// maximum size of message buffer 
        NULL  	// address of array of message inserts 
    );

    ErrorOk(prefix); 
    ErrorHandler(lpMsgBuf);
    LocalFree(lpMsgBuf);    // ???
} //SysErrorMessage
#endif

static void *alloc(size_t n) {
    void *p = calloc(n, 1);
    if (!p) {
        ErrorHandler("Not enough memory (calloc failed)");
		//should do a global error return here?? use longjmp?
    }
    return p;
} //alloc

static void put(const void *buffer, size_t n) {
    size_t count = fwrite(buffer, 1, n, fOut);
    if (count != n) {
        //Info("n = %d, count = %d\n", n, count);
        ErrorOk("Error writing file %s\n"
                "(fwrite failed)", szOutFileName);
    }
} //put

static void get(void *buffer, size_t n) {
    size_t count;

    lastFilePos = currentFilePos;
    if (bRamFile) {
        memcpy(buffer, pRamFile + currentFilePos, n);
        currentFilePos += n;
    }
    else {
        count = fread(buffer, 1, n, fExe);
        currentFilePos += count;
        if (count != n) {
            //Info("n = %d, count = %d\n", n, count);
            ErrorHandler("Invalid file format (fread failed)");
        }
    }
} //get

static void seekto(size_t offset) {
    if (!bRamFile) {
        if (_fseeki64(fExe, offset, SEEK_SET)) {
            ErrorHandler("Invalid file format (fseek failed)");
        }
    }
    currentFilePos = offset;
} //seekto

static void oseekto(size_t offset) {
    if (_fseeki64(fOut, offset, SEEK_SET)) {
        ErrorOk("Error writing file %s\n"
                "fseek failed", szOutFileName);
		//need error return???
    }
} //oseekto

static int findSection(UINT64 rva) {
    int i;
    DWORD sectionSize;

    if (pST) {
        /* Search the section table entries for the */
        /* section containing this RVA, if any...   */
        for (i = 0; i < nST; ++i) {
            sectionSize = max(pST[i].Misc.VirtualSize, pST[i].SizeOfRawData);
            if (rva >= pST[i].VirtualAddress &&
                rva < pST[i].VirtualAddress + sectionSize) {
                /* Found a section containing the RVA */
                return i;
            }
        }
    }
    return -1; /* not found */
} //findSection

#if 0
static int oFindSection(UINT64 rva) {
    int i;
    DWORD sectionSize;

    if (oST) {
        /* Search the output section table entries for */
        /* the section containing this RVA, if any...  */
        for (i = 0; i < nST; ++i) {
            sectionSize = max(oST[i].Misc.VirtualSize, oST[i].SizeOfRawData);
            if (rva >= oST[i].VirtualAddress &&
                rva < oST[i].VirtualAddress + sectionSize) {
                /* Found a section containing the RVA */
                return i;
            }
        }
    }
    return -1; /* not found */
} //oFindSection
#endif

static DWORD relocateOutputRVA(DWORD rva)
{
    // rva == 0 means not present
    // but note that rva != 0 below is really redundant!!!
    if (rva != 0 && rva >= topRsrcRVA) {
        return rva + outputRVAdelta;
    }
    else {
        return rva;
    }
} //relocateOutputRVA

static DWORD relocateOutputFileOffset(DWORD offset)
{
    // offset == 0 means not present
    // but note that offset != 0 below is really redundant!!!
    if (offset != 0 && offset >= topRsrcFileOffset) {
        return offset + outputFileOffsetDelta;
    }
    else {
        return offset;
    }
} //relocateOutputFileOffset

static size_t FileOffsetFromRVA(DWORD rva) {
    int i;

    i = findSection(rva);
    if (i >= 0) {
        /* Found section containing this RVA ...  */
        /* But does initialized data for this RVA */
        /* exist within the file, or is it BSS?   */
        if (rva < pST[i].VirtualAddress + pST[i].SizeOfRawData) {
            /* Valid seek address within file... */
            return pST[i].PointerToRawData + (rva - pST[i].VirtualAddress);
        }
        else {
            /* Not a valid seek address... */
            ErrorHandler("Invalid file format (seek to BSS in FileOffsetFromRVA)");
            return (size_t)-1;
        }
    }
    else {
        /* Not a valid RVA??? */
        ErrorHandler("Invalid file format (FileOffsetFromRVA failed)");
        return (size_t)-1;
    }
} //FileOffsetFromRVA

#if 0
static size_t oFileOffsetFromRVA(DWORD rva) {
    int i;

    i = oFindSection(rva);
    if (i >= 0) {
        /* Found section containing this RVA ...  */
        /* But does initialized data for this RVA */
        /* exist within the file, or is it BSS?   */
        if (rva < oST[i].VirtualAddress + oST[i].SizeOfRawData) {
            /* Valid seek address within file... */
            return oST[i].PointerToRawData + (rva - oST[i].VirtualAddress);
        }
        else {
            /* Not a valid seek address... */
            ErrorHandler("Invalid file format (seek to BSS in oFileOffsetFromRVA)");
            return (size_t)-1;
        }
    }
    else {
        /* Not a valid RVA??? */
        ErrorHandler("Invalid file format (oFileOffsetFromRVA failed)");
        return (size_t)-1;
    }
} //oFileOffsetFromRVA
#endif

static void seektoRVA(DWORD rva) {
    seekto(FileOffsetFromRVA(rva));
} //seektoRVA

#if 0
static void oseektoRVA(DWORD rva) {
    oseekto(oFileOffsetFromRVA(rva));
} //oseektoRVA
#endif


/////////////////////   Resource Scan Routines   ////////////////////////////


#if 0
// From <WinNT.h>, replicated here for documentation only!!!

typedef struct _IMAGE_RESOURCE_DIRECTORY {
    DWORD   Characteristics;
    DWORD   TimeDateStamp;
    WORD    MajorVersion;
    WORD    MinorVersion;
    WORD    NumberOfNamedEntries;
    WORD    NumberOfIdEntries;
//  IMAGE_RESOURCE_DIRECTORY_ENTRY DirectoryEntries[];
} IMAGE_RESOURCE_DIRECTORY, *PIMAGE_RESOURCE_DIRECTORY;

typedef struct _IMAGE_RESOURCE_DIRECTORY_ENTRY {
    union {
        struct {
            DWORD NameOffset:31;
            DWORD NameIsString:1;
        };
        DWORD   Name;
        WORD    Id;
    };
    union {
        DWORD   OffsetToData;
        struct {
            DWORD   OffsetToDirectory:31;
            DWORD   DataIsDirectory:1;
        };
    };
} IMAGE_RESOURCE_DIRECTORY_ENTRY, *PIMAGE_RESOURCE_DIRECTORY_ENTRY;

typedef struct _IMAGE_RESOURCE_DIR_STRING_U {
    WORD    Length;
    WCHAR   NameString[ 1 ];
} IMAGE_RESOURCE_DIR_STRING_U, *PIMAGE_RESOURCE_DIR_STRING_U;

typedef struct _IMAGE_RESOURCE_DATA_ENTRY {
    DWORD   OffsetToData;
    DWORD   Size;
    DWORD   CodePage;
    DWORD   Reserved;
} IMAGE_RESOURCE_DATA_ENTRY, *PIMAGE_RESOURCE_DATA_ENTRY;

#endif  //from <WinNT.h>

typedef struct _RES_DATA_NODE {
    DWORD NodeOffset;
    void *pActualData;
    IMAGE_RESOURCE_DATA_ENTRY embedded;
} RES_DATA_NODE;

typedef struct _RES_NAME_NODE {
    DWORD NodeOffset;
    struct _RES_NAME_NODE *pSameHashLink;
    BOOL bAlreadyVisited;
    IMAGE_RESOURCE_DIR_STRING_U embedded;    //variable length
} RES_NAME_NODE;

struct _RES_DIR_NODE;   //forward reference

typedef struct _RES_ENTRY_NODE {
    DWORD NodeOffset;
    struct _RES_ENTRY_NODE *pNextEntryNode;
    RES_NAME_NODE *pResName;            //if NameIsString
    union {
        struct _RES_DIR_NODE *pSubDir;  //if DataIsDirectory
        RES_DATA_NODE *pDataNode;       //if !DataIsDirectory
    };
    IMAGE_RESOURCE_DIRECTORY_ENTRY embedded;
} RES_ENTRY_NODE;

typedef struct _RES_DIR_NODE {
    DWORD NodeOffset;
    RES_ENTRY_NODE *pNamedEntryList;
    RES_ENTRY_NODE *pIdEntryList;
    IMAGE_RESOURCE_DIRECTORY embedded;
} RES_DIR_NODE;

static RES_DIR_NODE *pResDirRoot = NULL;
static RES_NAME_NODE *pNameNodeHashList = NULL;

static int NextDirNodeOffset = 0;      //DWORD aligned
static int NextDataNodeOffset = 0;     //DWORD aligned
static int NextNameNodeOffset = 0;     //WORD aligned
static int NextActualDataOffset = 0;   //DWORD aligned
static int NewResourceSize = 0;        //DWORD aligned

static int FirstDirNodeOffset = 0;      //DWORD aligned
static int FirstDataNodeOffset = 0;     //DWORD aligned
static int FirstNameNodeOffset = 0;     //WORD aligned
static int FirstActualDataOffset = 0;   //Paragraph aligned???

#define NEW(p)  \
    (p) = alloc(sizeof(*(p)))

#define DISPOSE(p)  \
    free(p)

static void seekRes(DWORD offset) {
    /* Seek to specified offset within resource info */
    seekto(ResRootOffset + offset);
} //seekRes

static void oseekRes(DWORD offset) {
    /* Seek to specified offset within resource info in output file */
    oseekto(ResRootOffset + offset);
} //oseekRes

static RES_NAME_NODE *DumpResString(const char *name, DWORD offset) {
    IMAGE_RESOURCE_DIR_STRING_U resString;
    //PIMAGE_RESOURCE_DIR_STRING_U pResString;
    RES_NAME_NODE *pNameNode = NULL;
    size_t saveFilePos = currentFilePos;

    /* Get enough of the string to determine its length... */
    seekRes(offset);
    get(&resString, sizeof(resString));
    /* Now allocate memory for it, and backup to get the entire string... */
    //pResString = alloc(resString.Length*sizeof(WCHAR) + sizeof(resString.Length));
    pNameNode = alloc(sizeof(*pNameNode) - sizeof(pNameNode->embedded.NameString[0]) +
        resString.Length*sizeof(WCHAR));
    seekRes(offset);
    get(&pNameNode->embedded, resString.Length*sizeof(WCHAR) + sizeof(resString.Length));
    {//scope
        /* Look for an already existing identical string... */
        RES_NAME_NODE *p;
        for (p = pNameNodeHashList; p != NULL; p = p->pSameHashLink) {
            if (p->embedded.Length == pNameNode->embedded.Length &&
                0 == memcmp(p->embedded.NameString, pNameNode->embedded.NameString,
                            p->embedded.Length*sizeof(WCHAR))) break;
        }
        if (p != NULL) {
            /* Found a match -- use it */
            DISPOSE(pNameNode);
            pNameNode = p;
        }
        else {
            /* New string.  Add it to the hash list */
            pNameNode->pSameHashLink = pNameNodeHashList;
            pNameNodeHashList = pNameNode;
            pNameNode->bAlreadyVisited = FALSE;
            pNameNode->NodeOffset = NextNameNodeOffset;
            NextNameNodeOffset += pNameNode->embedded.Length*sizeof(WCHAR) +
                                  sizeof(pNameNode->embedded.Length);   //WORD aligned
        }
    }//scope
    //restore prior file position...
    seekto(saveFilePos);
    return pNameNode;
} //DumpResString

static RES_DATA_NODE *DumpResData(DWORD offset) {
    IMAGE_RESOURCE_DATA_ENTRY resDataEntry;
    RES_DATA_NODE *pDataNode = NULL;
    size_t saveFilePos = currentFilePos;

    seekRes(offset);
    get(&resDataEntry, sizeof(resDataEntry));
    {//scope
        /* Scan the raw resource data... */
        void *p;

        seektoRVA(resDataEntry.OffsetToData);
        p = alloc(resDataEntry.Size);
        get(p, resDataEntry.Size);
        NEW(pDataNode);
        pDataNode->embedded = resDataEntry;
        pDataNode->pActualData = p;
        pDataNode->embedded.OffsetToData = NextActualDataOffset;
        NextActualDataOffset += (resDataEntry.Size + 3) & ~3;       //DWORD aligned
        pDataNode->NodeOffset = NextDataNodeOffset;
        NextDataNodeOffset += sizeof(IMAGE_RESOURCE_DATA_ENTRY);    //DWORD aligned
    }//scope
    //restore prior file position...
    seekto(saveFilePos);
    return pDataNode;
} //DumpResData

static RES_DIR_NODE *DumpResDir(DWORD offset) {
    /* Recursively scan the resource directory... */
    IMAGE_RESOURCE_DIRECTORY resDir;
    IMAGE_RESOURCE_DIRECTORY_ENTRY resDirEntry;
    RES_DIR_NODE *pDirNode = NULL;
    RES_ENTRY_NODE *pEntryNode;
    RES_NAME_NODE *pNameNode;
    void *pSubNode;
    int i;
    size_t saveFilePos = currentFilePos;

    RESLEVEL++;

    seekRes(offset);
    get(&resDir, sizeof(resDir));
    NEW(pDirNode);
    pDirNode->embedded = resDir;
    pDirNode->pNamedEntryList = NULL;
    pDirNode->pIdEntryList = NULL;
    pDirNode->NodeOffset = NextDirNodeOffset;
    NextDirNodeOffset += sizeof(IMAGE_RESOURCE_DIRECTORY) +
        (resDir.NumberOfNamedEntries + resDir.NumberOfIdEntries)*
        sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //DWORD aligned
    /* Scan named directory entries... */
    for (i = 0; i < resDir.NumberOfNamedEntries; ++i) {
        get(&resDirEntry, sizeof(resDirEntry));
        pNameNode = DumpResString("Name offset", resDirEntry.NameOffset);
        if (resDirEntry.DataIsDirectory) {
            pSubNode = DumpResDir(resDirEntry.OffsetToDirectory);
        }
        else {
            pSubNode = DumpResData(resDirEntry.OffsetToData);
        }
        NEW(pEntryNode);
        /* Force correct value for NameIsString */
        resDirEntry.NameIsString = TRUE;
        pEntryNode->embedded = resDirEntry;
        pEntryNode->pNextEntryNode = pDirNode->pNamedEntryList;
        pDirNode->pNamedEntryList = pEntryNode;
        pEntryNode->pResName = pNameNode;
        pEntryNode->pSubDir = pSubNode;
        pEntryNode->NodeOffset = pDirNode->NodeOffset +
                                 sizeof(IMAGE_RESOURCE_DIRECTORY) +
                                 sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY)*i;
    }//for named entries 
    /* Scan ID directory entries... */
    for (i = 0; i < resDir.NumberOfIdEntries; ++i) {
        get(&resDirEntry, sizeof(resDirEntry));
        if (resDirEntry.DataIsDirectory) {
            pSubNode = DumpResDir(resDirEntry.OffsetToDirectory);
        }
        else {
            pSubNode = DumpResData(resDirEntry.OffsetToData);
        }
        NEW(pEntryNode);
        /* Force correct value for NameIsString */
        resDirEntry.NameIsString = FALSE;
        pEntryNode->embedded = resDirEntry;
        pEntryNode->pNextEntryNode = pDirNode->pIdEntryList;
        pDirNode->pIdEntryList = pEntryNode;
        pEntryNode->pResName = NULL;
        pEntryNode->pSubDir = pSubNode;
        pEntryNode->NodeOffset = pDirNode->NodeOffset +
                                 sizeof(IMAGE_RESOURCE_DIRECTORY) +
                                 sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY)*
                                 (resDir.NumberOfNamedEntries + i);
    }//for ID entries
    {//scope
        /* Now reverse the entry chains... */
        RES_ENTRY_NODE *prevEntryNode;
        RES_ENTRY_NODE *nextEntryNode;

        /* Reverse the named entry chain */
        prevEntryNode = NULL;
        for (pEntryNode = pDirNode->pNamedEntryList;
             pEntryNode != NULL;
             pEntryNode = nextEntryNode)
        {
            nextEntryNode = pEntryNode->pNextEntryNode;
            pEntryNode->pNextEntryNode = prevEntryNode;
            prevEntryNode = pEntryNode;
        }//for pEntryNode
        pDirNode->pNamedEntryList = prevEntryNode;

        /* Reverse the ID entry chain */
        prevEntryNode = NULL;
        for (pEntryNode = pDirNode->pIdEntryList;
             pEntryNode != NULL;
             pEntryNode = nextEntryNode)
        {
            nextEntryNode = pEntryNode->pNextEntryNode;
            pEntryNode->pNextEntryNode = prevEntryNode;
            prevEntryNode = pEntryNode;
        }//for pEntryNode
        pDirNode->pIdEntryList = prevEntryNode;
    }//scope

    //restore prior file position...
    seekto(saveFilePos);
    RESLEVEL--;
    return pDirNode;
} //DumpResDir


static void resetResNameNodes(RES_DIR_NODE *pDirNode) {
    /* Recursively reset the resource name string nodes */
    RES_ENTRY_NODE *pEntryNode;

    /* Scan named subdirectory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        /* Reset the "already visited" flag */
        pEntryNode->pResName->bAlreadyVisited = FALSE;
        if (pEntryNode->embedded.DataIsDirectory) {
            resetResNameNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode
    /* Scan ID subdirectory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            resetResNameNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode
} //resetResNameNodes


static void assignResDirNodes(RES_DIR_NODE *pDirNode) {
    /* Recursively assign offsets for the resource directory nodes */
    RES_ENTRY_NODE *pEntryNode;

    /* Assign offset for this directory */
    pDirNode->NodeOffset = NextDirNodeOffset;
    NextDirNodeOffset += sizeof(IMAGE_RESOURCE_DIRECTORY); //DWORD aligned

    /* Assign named directory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        pEntryNode->NodeOffset = NextDirNodeOffset;
        NextDirNodeOffset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //DWORD aligned
    }//for pEntryNode
    /* Assign ID directory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        pEntryNode->NodeOffset = NextDirNodeOffset;
        NextDirNodeOffset += sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY); //DWORD aligned
    }//for pEntryNode
    /* Now recursively assign subdirectory nodes... */
    /* Assign named subdirectory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            assignResDirNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode
    /* Assign ID subdirectory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            assignResDirNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode
} //assignResDirNodes


static void assignResDataNodes(RES_DIR_NODE *pDirNode) {
    /* Recursively assign offsets for the resource data descriptor nodes */
    RES_ENTRY_NODE *pEntryNode;

    /* Scan named subdirectory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            assignResDataNodes(pEntryNode->pSubDir);
        }
        else {
            pEntryNode->pDataNode->embedded.OffsetToData = NextActualDataOffset;
            NextActualDataOffset += (pEntryNode->pDataNode->embedded.Size + 3) & ~3; //DWORD aligned
            pEntryNode->pDataNode->NodeOffset = NextDataNodeOffset;
            NextDataNodeOffset += sizeof(IMAGE_RESOURCE_DATA_ENTRY);    //DWORD aligned
        }
    }//for pEntryNode
    /* Scan ID subdirectory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            assignResDataNodes(pEntryNode->pSubDir);
        }
        else {
            pEntryNode->pDataNode->embedded.OffsetToData = NextActualDataOffset;
            NextActualDataOffset += (pEntryNode->pDataNode->embedded.Size + 3) & ~3; //DWORD aligned
            pEntryNode->pDataNode->NodeOffset = NextDataNodeOffset;
            NextDataNodeOffset += sizeof(IMAGE_RESOURCE_DATA_ENTRY);    //DWORD aligned
        }
    }//for pEntryNode
} //assignResDataNodes


static void assignResNameNodes(RES_DIR_NODE *pDirNode) {
    /* Recursively assign offsets for the resource name string nodes */
    RES_ENTRY_NODE *pEntryNode;

    /* Scan named subdirectory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        RES_NAME_NODE *pNameNode = pEntryNode->pResName;

        if (!pNameNode->bAlreadyVisited) {
            pNameNode->NodeOffset = NextNameNodeOffset;
            NextNameNodeOffset += pNameNode->embedded.Length*sizeof(WCHAR) +
                                  sizeof(pNameNode->embedded.Length);   //WORD aligned
            pNameNode->bAlreadyVisited = TRUE;   //set "already assigned" flag
        }
        if (pEntryNode->embedded.DataIsDirectory) {
            assignResNameNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode
    /* Scan ID subdirectory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            assignResNameNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode
} //assignResNameNodes


static void assignResources(RES_DIR_NODE *pResDirRoot)
{
    DWORD newSectionAlignedSize;
    DWORD oldSectionAlignedSize;
    DWORD newFileAlignedSize;
    DWORD oldFileAlignedSize;
    int i;

    //(Re-)assign all node offsets

    /* Reset all beginning node offsets */
    NextDirNodeOffset = 0;
    NextDataNodeOffset = 0;
    NextNameNodeOffset = 0;
    NextActualDataOffset = 0;

    /* Clear "already visited" flags */
    resetResNameNodes(pResDirRoot);

    assignResDirNodes(pResDirRoot);
    assignResDataNodes(pResDirRoot);
    assignResNameNodes(pResDirRoot);
    //assignResActualData(pResDirRoot);

    /* Adjust base offsets... */
    FirstDirNodeOffset    =  0;
    FirstDataNodeOffset   =  FirstDirNodeOffset + NextDirNodeOffset;
    FirstNameNodeOffset   =  FirstDataNodeOffset + NextDataNodeOffset;
    //Paragraph-align the start of actual data...
    FirstActualDataOffset = (FirstNameNodeOffset + NextNameNodeOffset + 15) & ~15;
    NewResourceSize = FirstActualDataOffset + NextActualDataOffset; //dword aligned

    /* Compute old and new file-aligned size of resources */
    oldFileAlignedSize = pST[rsrcSect].SizeOfRawData;
    newFileAlignedSize =
        newOptHeader.FileAlignment != 0 ?
        ((NewResourceSize + newOptHeader.FileAlignment - 1)/newOptHeader.FileAlignment)*
        newOptHeader.FileAlignment :
        NewResourceSize;

    /* Compute old and new section-aligned size of resources    */
    /* (Don't use old Misc.VirtualSize -- it's not always set!  */
    /* But whoops -- the file-aligned size might include just   */
    /* the initialized portion of the data, not the full VM     */
    /* size!!!) */
    oldSectionAlignedSize = 
        optHeader.SectionAlignment != 0 ?
        ((oldFileAlignedSize + optHeader.SectionAlignment - 1)/optHeader.SectionAlignment)*
                                                               optHeader.SectionAlignment :
        oldFileAlignedSize;
    newSectionAlignedSize =
        newOptHeader.SectionAlignment != 0 ?
        ((NewResourceSize + newOptHeader.SectionAlignment - 1)/newOptHeader.SectionAlignment)*
                                                               newOptHeader.SectionAlignment :
        NewResourceSize;

    topRsrcRVA = pST[rsrcSect].VirtualAddress + oldSectionAlignedSize;
    outputRVAdelta = newSectionAlignedSize - oldSectionAlignedSize;

    topRsrcFileOffset = pST[rsrcSect].PointerToRawData + oldFileAlignedSize;
    outputFileOffsetDelta = newFileAlignedSize - oldFileAlignedSize;

    /* Update section table for the output resource section */
    oST[rsrcSect].Misc.VirtualSize = NewResourceSize;
    oST[rsrcSect].SizeOfRawData = newFileAlignedSize;

#if 0
    putd32("Old size of resources", pST[rsrcSect].Misc.VirtualSize);
    putd32("Old file-aligned size", pST[rsrcSect].SizeOfRawData);
    putd32("Old section-aligned size", oldSectionAlignedSize);
    putd32("New size of resources", oST[rsrcSect].Misc.VirtualSize);
    putd32("New file-aligned size", oST[rsrcSect].SizeOfRawData);
    putd32("New section-aligned size", newSectionAlignedSize);
    putd32("Top .rsrc RVA", topRsrcRVA);
    putd32("RVA delta", outputRVAdelta);
    putd32("Top .rsrc file offset", topRsrcFileOffset);
    putd32("File offset delta", outputFileOffsetDelta);
#endif

    /* Update RVA's and file offsets in the output Section Table */
    for (i = 0; i < nST; ++i) {
        oST[i].VirtualAddress = relocateOutputRVA(pST[i].VirtualAddress);
        oST[i].PointerToRawData = relocateOutputFileOffset(pST[i].PointerToRawData);
        oST[i].PointerToRelocations = relocateOutputFileOffset(pST[i].PointerToRelocations);
        oST[i].PointerToLinenumbers = relocateOutputFileOffset(pST[i].PointerToLinenumbers);
    }
} //assignResources


static void deleteResDirNodes(RES_DIR_NODE *pDirNode)
{
    RES_ENTRY_NODE *pEntryNode;

    /* Delete named directory entries... */
    while ((pEntryNode = pDirNode->pNamedEntryList) != NULL)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            /* Recursively delete subdirectory */
            deleteResDirNodes(pEntryNode->pSubDir);
        }
        else {
            /* Delete data node and its actual data */
            DISPOSE(pEntryNode->pDataNode->pActualData);
            DISPOSE(pEntryNode->pDataNode);
        }
        pDirNode->pNamedEntryList = pEntryNode->pNextEntryNode;
        DISPOSE(pEntryNode);
    }//while pEntryNode
    /* Delete ID directory entries... */
    while ((pEntryNode = pDirNode->pIdEntryList) != NULL)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            /* Recursively delete subdirectory */
            deleteResDirNodes(pEntryNode->pSubDir);
        }
        else {
            /* Delete data node and its actual data */
            DISPOSE(pEntryNode->pDataNode->pActualData);
            DISPOSE(pEntryNode->pDataNode);
        }
        pDirNode->pIdEntryList = pEntryNode->pNextEntryNode;
        DISPOSE(pEntryNode);
    }//while pEntryNode
    /* Finally, delete the directory node itself */
    DISPOSE(pDirNode);
} //deleteResDirNodes


static void deleteResNodes(void)
{
    //Delete the resource tree

    if (pResDirRoot) {									//fix this back in AfhPED!!!??
        /* Recursively delete all (sub)directories */
        deleteResDirNodes(pResDirRoot);
        pResDirRoot = NULL;     //it's gone!
    }

    /* Now delete the name nodes by following the hash chain(s) */
    while (pNameNodeHashList != NULL) {
        RES_NAME_NODE *p = pNameNodeHashList;
        pNameNodeHashList = pNameNodeHashList->pSameHashLink;
        DISPOSE(p);
    }//while

} //deleteResNodes


static void writeResDirNodes(RES_DIR_NODE *pDirNode) {
    /* Recursively write the resource directory nodes */
    RES_ENTRY_NODE *pEntryNode;

    RESLEVEL++;

    /* Write directory header to output file */
    oseekRes(pDirNode->NodeOffset + FirstDirNodeOffset);
    put(&pDirNode->embedded, sizeof(pDirNode->embedded));
    /* Write named directory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        /* Update embedded offset fields */
        pEntryNode->embedded.NameOffset = pEntryNode->pResName->NodeOffset +
                                          FirstNameNodeOffset;
        if (pEntryNode->embedded.DataIsDirectory) {
            pEntryNode->embedded.OffsetToDirectory = pEntryNode->pSubDir->NodeOffset +
                                                     FirstDirNodeOffset;
        }
        else {
            pEntryNode->embedded.OffsetToData = pEntryNode->pDataNode->NodeOffset +
                                                FirstDataNodeOffset;
        }
        /* Write named directory entry to output file */
        oseekRes(pEntryNode->NodeOffset + FirstDirNodeOffset);  //redundant!!!
        put(&pEntryNode->embedded, sizeof(pEntryNode->embedded));
    }//for pEntryNode
    /* Write ID directory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        /* Update embedded offset fields */
        if (pEntryNode->embedded.DataIsDirectory) {
            pEntryNode->embedded.OffsetToDirectory = pEntryNode->pSubDir->NodeOffset +
                                                     FirstDirNodeOffset;
        }
        else {
            pEntryNode->embedded.OffsetToData = pEntryNode->pDataNode->NodeOffset +
                                                FirstDataNodeOffset;
        }
        /* Write ID directory entry to output file */
        oseekRes(pEntryNode->NodeOffset + FirstDirNodeOffset);  //redundant!!!
        put(&pEntryNode->embedded, sizeof(pEntryNode->embedded));
    }//for pEntryNode
    /* Now recursively write out subdirectory nodes... */
    /* Write named subdirectory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            writeResDirNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode
    /* Write ID subdirectory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            writeResDirNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode

    RESLEVEL--;
} //writeResDirNodes

static void writeResDataNode(RES_DATA_NODE *pDataNode) {
    /* Write data node to output file */
    oseekRes(pDataNode->NodeOffset + FirstDataNodeOffset);
    /* Temporarily relocate the actual data offset to an RVA */
    pDataNode->embedded.OffsetToData += FirstActualDataOffset +
                                        oST[rsrcSect].VirtualAddress;
    put(&pDataNode->embedded, sizeof(pDataNode->embedded));
    /* Undo the temporary relocation */
    pDataNode->embedded.OffsetToData -= FirstActualDataOffset +
                                        oST[rsrcSect].VirtualAddress;
} //writeResDataNode

static void writeResDataNodes(RES_DIR_NODE *pDirNode) {
    /* Recursively write the resource data descriptor nodes */
    RES_ENTRY_NODE *pEntryNode;

    RESLEVEL++;

    /* Scan named subdirectory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            writeResDataNodes(pEntryNode->pSubDir);
        }
        else {
            writeResDataNode(pEntryNode->pDataNode);
        }
    }//for pEntryNode
    /* Scan ID subdirectory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            writeResDataNodes(pEntryNode->pSubDir);
        }
        else {
            writeResDataNode(pEntryNode->pDataNode);
        }
    }//for pEntryNode

    RESLEVEL--;
} //writeResDataNodes

static void writeResData(RES_DATA_NODE *pDataNode) {
    /* Write actual data to output file */
    oseekRes(pDataNode->embedded.OffsetToData + FirstActualDataOffset);
    put(pDataNode->pActualData, pDataNode->embedded.Size);
} //writeResData

static void writeResActualData(RES_DIR_NODE *pDirNode) {
    /* Recursively write the resource actual data */
    RES_ENTRY_NODE *pEntryNode;

    RESLEVEL++;

    /* Scan named subdirectory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            writeResActualData(pEntryNode->pSubDir);
        }
        else {
            writeResData(pEntryNode->pDataNode);
        }
    }//for pEntryNode
    /* Scan ID subdirectory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            writeResActualData(pEntryNode->pSubDir);
        }
        else {
            writeResData(pEntryNode->pDataNode);
        }
    }//for pEntryNode

    RESLEVEL--;
} //writeResActualData

static void writeResNameNodes(RES_DIR_NODE *pDirNode) {
    /* Recursively write the resource name string nodes */
    RES_ENTRY_NODE *pEntryNode;

    RESLEVEL++;

    /* Scan named subdirectory entries... */
    for (pEntryNode = pDirNode->pNamedEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        RES_NAME_NODE *pNameNode = pEntryNode->pResName;

        if (pNameNode->bAlreadyVisited) {
            //Info("**** DUPLICATE ****");
        }
        else {
            /* Write name string to output file */
            oseekRes(pNameNode->NodeOffset + FirstNameNodeOffset);
            put(&pNameNode->embedded,
                pNameNode->embedded.Length*sizeof(WCHAR) +
                sizeof(pNameNode->embedded.Length));
            pNameNode->bAlreadyVisited = TRUE;   //set "already written" flag
        }
        if (pEntryNode->embedded.DataIsDirectory) {
            writeResNameNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode
    /* Scan ID subdirectory entries... */
    for (pEntryNode = pDirNode->pIdEntryList;
         pEntryNode != NULL;
         pEntryNode = pEntryNode->pNextEntryNode)
    {
        if (pEntryNode->embedded.DataIsDirectory) {
            writeResNameNodes(pEntryNode->pSubDir);
        }
    }//for pEntryNode

    RESLEVEL--;
} //writeResNameNodes


static void writeResources(RES_DIR_NODE *pResDirRoot)
{
    /* Adjust base offsets... */
    FirstDirNodeOffset    =  0;
    FirstDataNodeOffset   =  FirstDirNodeOffset + NextDirNodeOffset;
    FirstNameNodeOffset   =  FirstDataNodeOffset + NextDataNodeOffset;
    //Paragraph-align the start of actual data...
    FirstActualDataOffset = (FirstNameNodeOffset + NextNameNodeOffset + 15) & ~15;
    NewResourceSize = FirstActualDataOffset + NextActualDataOffset; //dword aligned

    /* Clear "already visited" flags */
    resetResNameNodes(pResDirRoot);

    /* First write out resource directory nodes... */
    writeResDirNodes(pResDirRoot);

    /* Now write out resource data descriptor nodes... */
    writeResDataNodes(pResDirRoot);

    /* Next, write resource name string nodes... */
    writeResNameNodes(pResDirRoot);

    /* Finally, write the actual resource data... */
    writeResActualData(pResDirRoot);
}//writeResources


static RES_ENTRY_NODE *locateResDirEntry(RES_DIR_NODE *pDir, LPCWSTR lpResName)
{
    RES_ENTRY_NODE *pDirEntry;

    if (!pDir) return NULL;

/// if (((DWORD)lpResName & 0xFFFF0000) != 0) {
    if (((UINT_PTR)lpResName & ~0xFFFF) != 0) {		//put this fix in AfhPED???
        //it's a named resource
        for (pDirEntry = pDir->pNamedEntryList;
             pDirEntry != NULL;
             pDirEntry = pDirEntry->pNextEntryNode)
        {
            //(could stop when NameString >= lpResName, since list is sorted,
            // or could even do a binary search, but hardly worth it)
            if (pDirEntry->pResName->embedded.Length == wcslen(lpResName) &&
                _wcsnicmp(pDirEntry->pResName->embedded.NameString, lpResName,
                          pDirEntry->pResName->embedded.Length) == 0) {
                //found it
                return pDirEntry;
            }
        }//for
    }
    else {
        //it's an ordinal resource ID
        for (pDirEntry = pDir->pIdEntryList;
             pDirEntry != NULL;
             pDirEntry = pDirEntry->pNextEntryNode)
        {
            //(could stop when Id >= lpResName, since list is sorted,
            // or could even do a binary search, but hardly worth it)
            if (pDirEntry->embedded.Id == (WORD)lpResName) {
                //found it
                return pDirEntry;
            }
        }//for
    }
    return NULL;        //not found
} //locateResDirEntry


static RES_NAME_NODE *enterResName(LPCWSTR lpResName)
{
    RES_NAME_NODE *pNameNode;
    RES_NAME_NODE *p;
    WORD wLength = (WORD)wcslen(lpResName);

    pNameNode = alloc(sizeof(*pNameNode) - sizeof(pNameNode->embedded.NameString[0]) +
                      wLength*sizeof(WCHAR));
    pNameNode->NodeOffset = 0;
    pNameNode->pSameHashLink = NULL;
    pNameNode->bAlreadyVisited = FALSE;
    pNameNode->embedded.Length = wLength;
    wcsncpy(pNameNode->embedded.NameString, lpResName, wLength);
    //convert resource name to uppercase for case-insensitivity
#if 0														//put this fix in AfhPED???
	//use secure version of _wcsupr()
    _wcsupr_s(pNameNode->embedded.NameString, wLength);	//NFG!! Namestring is not null-terminated!!!
#else
	//workaround if secure _wcsupr_s() is not available...
	{//scope
		int i;
        for (i = 0; i < wLength; i++) {
			pNameNode->embedded.NameString[i] = 
				towupper(pNameNode->embedded.NameString[i]);
        }//for
    }//scope
#endif

    /* Look for an already existing identical string... */
    for (p = pNameNodeHashList; p != NULL; p = p->pSameHashLink) {
        if (p->embedded.Length == wLength &&
            0 == memcmp(p->embedded.NameString, pNameNode->embedded.NameString,
                        wLength*sizeof(WCHAR))) break;
    }
    if (p != NULL) {
        /* Found a match -- use it */
        DISPOSE(pNameNode);
        pNameNode = p;
    }
    else {
        /* New string.  Add it to the hash list */
        pNameNode->pSameHashLink = pNameNodeHashList;
        pNameNodeHashList = pNameNode;
        pNameNode->NodeOffset = NextNameNodeOffset;
        NextNameNodeOffset += wLength*sizeof(WCHAR) +
                              sizeof(pNameNode->embedded.Length);   //WORD aligned
    }
    return pNameNode;
} //enterResName


static RES_ENTRY_NODE *insertResDirEntry(RES_DIR_NODE *pDir, LPCWSTR lpResName)
{
    //assumes lpResName is not already present in this directory;
    //otherwise, a duplicate entry will be inserted!!!
    RES_ENTRY_NODE *pDirEntry, *p, *q;

    if (!pDir) return NULL;

    NEW(pDirEntry);
    pDirEntry->NodeOffset = 0;
    pDirEntry->pNextEntryNode = NULL;
    pDirEntry->pResName = NULL;
    pDirEntry->pSubDir = NULL;
    pDirEntry->embedded.Name = 0;
    pDirEntry->embedded.OffsetToData = 0;

    if (((UINT_PTR)lpResName & ~0xFFFF) != 0) {		//put this fix in AfhPED???
        //it's a named resource
        pDir->embedded.NumberOfNamedEntries++;
        pDirEntry->embedded.NameIsString = TRUE;
        pDirEntry->pResName = enterResName(lpResName);
        q = NULL;   //trailing pointer
        for (p = pDir->pNamedEntryList; p != NULL; p = p->pNextEntryNode)
        {
            if (_wcsnicmp(p->pResName->embedded.NameString, lpResName,
                          p->pResName->embedded.Length) > 0 ||
                _wcsnicmp(p->pResName->embedded.NameString, lpResName,
                          p->pResName->embedded.Length) == 0 &&
                p->pResName->embedded.Length >= wcslen(lpResName)
               ) {
                //new entry goes here
                break;
            }
            q = p;
        }//for
        if (q) {
            //insert after q
            q->pNextEntryNode = pDirEntry;
        }
        else {
            //insert at head of named entry list
            pDir->pNamedEntryList = pDirEntry;
        }
    }
    else {
        //it's an ordinal resource ID
        pDir->embedded.NumberOfIdEntries++;
        pDirEntry->embedded.NameIsString = FALSE;
        pDirEntry->embedded.Id = (WORD)lpResName;
        q = NULL;   //trailing pointer
        for (p = pDir->pIdEntryList; p != NULL; p = p->pNextEntryNode)
        {
            if (p->embedded.Id >= (WORD)lpResName) {
                //new entry goes here
                break;
            }
            q = p;
        }//for
        if (q) {
            //insert after q
            q->pNextEntryNode = pDirEntry;
        }
        else {
            //insert at head of ID entry list
            pDir->pIdEntryList = pDirEntry;
        }
    }
    pDirEntry->pNextEntryNode = p;
    return pDirEntry;
} //insertResDirEntry


static void deleteResDirEntry(RES_DIR_NODE *pDir, RES_ENTRY_NODE *pDirEntry)
{
    RES_ENTRY_NODE *p, *q;

    if (pDirEntry->embedded.NameIsString) {
        //search named entry list
        //delete the name node also??? -- no, it may be used elsewhere,
        //so garbage collect it								//DOES IT EVER GARBAGE COLLECT???
        q = NULL;   //trailing pointer
        for (p = pDir->pNamedEntryList; p != NULL; p = p->pNextEntryNode)
        {
            if (p == pDirEntry) {
                //found it
                break;
            }
            q = p;
        }//for
        if (p) {
            if (q) {
                //pDirEntry follows q
                q->pNextEntryNode = pDirEntry->pNextEntryNode;
            }
            else {
                //delete pDirEntry from head of named entry list
                pDir->pNamedEntryList = pDirEntry->pNextEntryNode;
            }
            pDir->embedded.NumberOfNamedEntries--;
        }
        else {
            //very strange: pDirEntry was not found!!!	//SHOULD REPORT THIS, at least in AfhPED???
            //just delete pDirEntry and return silently
        }
    }
    else {
        //search ID entry list
        q = NULL;   //trailing pointer
        for (p = pDir->pIdEntryList; p != NULL; p = p->pNextEntryNode)
        {
            if (p == pDirEntry) {
                //found it
                break;
            q = p;
            }
        }//for
        if (p) {
            if (q) {
                //pDirEntry follows q
                q->pNextEntryNode = pDirEntry->pNextEntryNode;
            }
            else {
                //delete pDirEntry from head of ID entry list
                pDir->pIdEntryList = pDirEntry->pNextEntryNode;
            }
            pDir->embedded.NumberOfIdEntries--;
        }
        else {
            //very strange: pDirEntry was not found!!!	//SHOULD REPORT THIS, at least in AfhPED???
            //just delete pDirEntry and return silently
        }
    }
    DISPOSE(pDirEntry);
} //deleteResDirEntry


static RES_DIR_NODE *newResDirNode(void)
{
    //create an empty resource directory node
    RES_DIR_NODE *pDirNode;

    NEW(pDirNode);
    pDirNode->NodeOffset = 0;
    pDirNode->pNamedEntryList = NULL;
    pDirNode->pIdEntryList = NULL;
    pDirNode->embedded.Characteristics = 0;
    pDirNode->embedded.TimeDateStamp = (DWORD)time(NULL);
    pDirNode->embedded.MajorVersion = 0;
    pDirNode->embedded.MinorVersion = 0;
    pDirNode->embedded.NumberOfNamedEntries = 0;
    pDirNode->embedded.NumberOfIdEntries = 0;
    return pDirNode;
} //newResDirNode


static RES_DATA_NODE *newResDataNode(void)
{
    RES_DATA_NODE *pDataNode;

    NEW(pDataNode);
    pDataNode->NodeOffset = 0;
    pDataNode->pActualData = NULL;
    pDataNode->embedded.OffsetToData = 0;
    pDataNode->embedded.Size = 0;
    pDataNode->embedded.CodePage = CP_ACP;  //default to ANSI code page
    pDataNode->embedded.Reserved = 0;
    return pDataNode;
} //newResDataNode

static LPWSTR ResNameString(LPCTSTR lp)
{
    LPWSTR lpwc;
    int nChars;
    int nChars2;

    if (((UINT_PTR)lp & ~0xFFFF) != 0) {
        //it's a named resource string
        //convert it to Unicode
        //- first, compute required size of Unicode buffer
        nChars = MultiByteToWideChar(
            CP_ACP,	// ANSI code page 
            0,	    // default character-type options 
            lp,	    // address of string to map 
            -1,	    // compute number of characters in NUL-terminated string 
            NULL,	// address of wide-character buffer 
            0 	    // compute size of buffer (including final NUL)
           );
        if (nChars) {
            lpwc = alloc(nChars*sizeof(WCHAR));
            nChars2 = MultiByteToWideChar(
                CP_ACP,	// ANSI code page 
                0,	    // default character-type options 
                lp,	    // address of string to map 
                -1,	    // compute number of characters in NUL-terminated string 
                lpwc,	// address of wide-character buffer 
                nChars  // size of buffer (including final NUL)
               );
            if (nChars2 == nChars) {
                //successful conversion to Unicode
                return lpwc;
            }
            else {
                //conversion to Unicode failed
                free(lpwc);
                return NULL;
            }
        }
        else {
            //conversion to Unicode failed
            return NULL;
        }
    }
    else {
        //it's an integer resource id
        return (LPWSTR)lp;
    }
} //ResNameString

static void freeResNameString(LPWSTR lpResName)
{
    if (((UINT_PTR)lpResName & ~0xFFFF) != 0) {
        //it's a named resource string
        free(lpResName);
    }
} //freeResNameString

BOOL afhUpdateResource(
    HANDLE pResDir,	        // root of resource tree to update 
    LPCTSTR lpType,	        // address of resource type to update 
    LPCTSTR lpName,	        // address of resource name to update  
    WORD wLanguage,	        // language identifier of resource 
    LPVOID lpData,	        // address of resource data 
    DWORD cbData 	        // length of resource data, in bytes 
   )
{
    BOOL bDeleting = (lpData == NULL);  // lpData == NULL => deleting this resource
    RES_DIR_NODE *pDir1, *pDir2, *pDir3;
    RES_ENTRY_NODE *pDirEntry1, *pDirEntry2, *pDirEntry3;
    LPWSTR lpwcType;        // wide-char version of lpType
    LPWSTR lpwcName;        // wide-char version of lpName

    //convert ASCII names to Unicode
    lpwcType = ResNameString(lpType);
    lpwcName = ResNameString(lpName);

    //locate the resource type
    pDir1 = pResDir;
    pDirEntry1 = locateResDirEntry(pDir1, lpwcType);
    if (!pDirEntry1) {
        //type not found
        if (bDeleting) {
            //done if deleting
            freeResNameString(lpwcType);
            freeResNameString(lpwcName);
            return TRUE;    //success
        }
        //inserting: create a directory entry node for this resource type
        pDirEntry1 = insertResDirEntry(pDir1, lpwcType);
        pDirEntry1->embedded.DataIsDirectory = TRUE;
        pDirEntry1->pSubDir = newResDirNode();
    }
    else if (!pDirEntry1->embedded.DataIsDirectory) {
        //not a subdirectory entry
        freeResNameString(lpwcType);
        freeResNameString(lpwcName);
        return FALSE;       //failure
    }

    //locate the resource name
    pDir2 = pDirEntry1->pSubDir;
    pDirEntry2 = locateResDirEntry(pDir2, lpwcName);
    if (!pDirEntry2) {
        //name not found
        if (bDeleting) {
            //done if deleting
            freeResNameString(lpwcType);
            freeResNameString(lpwcName);
            return TRUE;    //success
        }
        //inserting: create a directory entry node for this resource name
        pDirEntry2 = insertResDirEntry(pDir2, lpwcName);
        pDirEntry2->embedded.DataIsDirectory = TRUE;
        pDirEntry2->pSubDir = newResDirNode();
    }
    else if (!pDirEntry2->embedded.DataIsDirectory) {
        //not a subdirectory entry
        freeResNameString(lpwcType);
        freeResNameString(lpwcName);
        return FALSE;       //failure
    }

    //locate the language ID
    pDir3 = pDirEntry2->pSubDir;
    pDirEntry3 = locateResDirEntry(pDir3, MAKEINTRESOURCEW(wLanguage));
    if (!pDirEntry3) {
        //language ID not found
        if (bDeleting) {
            //done if deleting
            freeResNameString(lpwcType);
            freeResNameString(lpwcName);
            return TRUE;    //success
        }
        //inserting: create a data entry node for this language ID
        pDirEntry3 = insertResDirEntry(pDir3, MAKEINTRESOURCEW(wLanguage));
        pDirEntry3->embedded.DataIsDirectory = FALSE;
        pDirEntry3->pDataNode = newResDataNode();
    }
    else if (pDirEntry3->embedded.DataIsDirectory) {
        //not a data entry
        freeResNameString(lpwcType);
        freeResNameString(lpwcName);
        return FALSE;       //failure
    }
    else {
        //delete the actual data belonging to this directory entry node
        DISPOSE(pDirEntry3->pDataNode->pActualData);
        if (bDeleting) {
            //deleting: delete the data node and directory entry node
            DISPOSE(pDirEntry3->pDataNode);
            deleteResDirEntry(pDir3, pDirEntry3);
            //could also check if upper level dir nodes are empty and can be
            //deleted, but probably not worth the trouble			//OR IS IT???
            freeResNameString(lpwcType);
            freeResNameString(lpwcName);
            return TRUE;    //success
        }
    }
    //allocate memory to hold the actual data
    pDirEntry3->pDataNode->NodeOffset = 0;
    pDirEntry3->pDataNode->pActualData = alloc(cbData);
    memcpy(pDirEntry3->pDataNode->pActualData, lpData, cbData);
    pDirEntry3->pDataNode->embedded.Size = cbData;
    pDirEntry3->pDataNode->embedded.OffsetToData = 0;   //for now (it's an RVA!!!)

    freeResNameString(lpwcType);
    freeResNameString(lpwcName);
    return TRUE;       //success

} //afhUpdateResource


HANDLE afhBeginUpdateResource(
    LPCTSTR pFileName,	            // pointer to file in which to update resources  
    void *pRamTemplate	            // deletion option or pointer to RAM template
   )
{
	WORD ExpectedOptHeaderMagic;

    /* Save desired file name */
    szFileName = pFileName;

    if ((UINT_PTR)pRamTemplate > 1) {
        // we have a pointer to a RAM file template
        bRamFile = TRUE;
        pRamFile = pRamTemplate;
    }
    else {
        // Arg 2 is just a "delete original resources" flag
        // (which is not currently implemented)
        bRamFile = FALSE;
    }

    if (!bRamFile) {
        // Open the disk file for input
        fExe = fopen(szFileName, "rb");
        if (!fExe) {
            ErrorHandler("file open failed");
            return NULL;    //failure
        }
    }

    lastFilePos = currentFilePos = 0;

    ////////////////// Preload a number of structures ////////////////

    /* We need to preload the following portions of the image */
    /* so we can decode RVA's early on...                     */

    /* Assume Data Directory and Section Table not available... */
	pDD = NULL;
	oDD = NULL;
    pST = NULL;
    oST = NULL;
    pFixups = NULL;
    nST = 0;

    /* Preload DOS header */
    DosHeaderOffset = currentFilePos;
    get(&dosHeader, sizeof(dosHeader));

    if (dosHeader.e_magic == IMAGE_DOS_SIGNATURE) {
        /* Preload PE signature */
        seekto(dosHeader.e_lfanew);
        get(&Signature, sizeof(Signature));

        if (Signature == IMAGE_NT_SIGNATURE) {
            /* Preload PE header */
            PeHeaderOffset = currentFilePos;
            get(&peHeader, sizeof(peHeader));

#if 1	//sanity check from AfhPED
			if (peHeader.SizeOfOptionalHeader > sizeof(optHeader)) {
				Warn("Invalid SizeOfOptionalHeader (%d.), greater than max (%d.)\n"
					 "Assuming SizeOfOptionalHeader = max valid size in order to continue\n",
					 peHeader.SizeOfOptionalHeader, sizeof(optHeader));
				peHeader.SizeOfOptionalHeader = sizeof(optHeader);
				///ErrorHandler("Invalid SizeOfOptionalHeader");
			}
#endif

			SectTableOffset = currentFilePos + peHeader.SizeOfOptionalHeader;
            nST = peHeader.NumberOfSections;

            /* Preload PE Optional header */
            PeOptHeaderOffset = currentFilePos;
            get(&optHeader, peHeader.SizeOfOptionalHeader);
            /* Save a copy for the output file */
            newOptHeader = optHeader;

#if 1	//from AfhPED 1.9
			/* Set the bPE32PLUS flag if this is a PE32+ file */
			bPE32PLUS = (optHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC);
			#define PE32PLUS (bPE32PLUS)	//only valid from hereon...

			/* Preload some needed fields, and set pointers to the Data Directories */
			if (PE32PLUS) {
				ImageBase = optHeader.u64.ImageBase;
				NumberOfRvaAndSizes = optHeader.u64.NumberOfRvaAndSizes;
				pDD = optHeader.u64.DataDirectory;
				oDD = newOptHeader.u64.DataDirectory;
				//Make sure we're using the correct template...
				if (!X64GEN) {
					Warn("Using wrong (64-bit) template to generate a 32-bit standalone!");
				}
			}
			else {//pe32
				ImageBase = optHeader.ImageBase;
				NumberOfRvaAndSizes = optHeader.NumberOfRvaAndSizes;
				pDD = optHeader.DataDirectory;
				oDD = newOptHeader.DataDirectory;
				//Make sure we're using the correct template...
				if (X64GEN) {
					Warn("Using wrong (32-bit) template to generate a 64-bit standalone!");
				}
			}

			if ((ImageBase & 0xffff) != 0) {
				/* Invalid ImageBase */
				Warn("ImageBase (0x%.8I64x) is not a multiple of 64K\n",
					 ImageBase);		//FIX THIS IN AfhPed!!
			}
#endif
            if (optHeader.Magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC ||
				optHeader.Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC) {
                /* Preload Section table */
                // Should we explicitly seek to SectTableOffset???
                pST = alloc(sizeof(IMAGE_SECTION_HEADER)*nST);
                get(pST, sizeof(IMAGE_SECTION_HEADER)*nST);
                /* Make a copy of the Section Table for the output file */
                oST = alloc(sizeof(IMAGE_SECTION_HEADER)*nST);
                memcpy(oST, pST, sizeof(IMAGE_SECTION_HEADER)*nST);
                /* Allocate (and initialize) our fixup info table */
                pFixups = alloc(sizeof(FIXUP_INFO)*nST);
#ifdef VERBOSE
                Info("Successfully loaded %d section table entries.\n", nST);
#endif
            }
        }
    }

    //////////////////   DOS HEADER   /////////////////////

    //get(&dosHeader, sizeof(dosHeader));

    if (dosHeader.e_magic != IMAGE_DOS_SIGNATURE) {
        ErrorHandler("Not a DOS executable image or DLL");
        free(pST);
        free(oST);
        free(pFixups);
        if (!bRamFile) {
            fclose(fExe);
        }
        return NULL;    //failure
    }

    //////////////////   DOS STUB   /////////////////////

    // DOS Stub program

    if (dosHeader.e_lfanew > dosHeader.e_lfarlc)
        stubSize = dosHeader.e_lfanew - dosHeader.e_lfarlc;
    else
        stubSize = 0;
    if (stubSize > 0 && dosHeader.e_lfarlc != 0) {
        seekto(dosHeader.e_lfarlc);
        pDosStub = alloc(stubSize);
        get(pDosStub, stubSize);
    }

    //////////////////   PE SIGNATURE   /////////////////////

    //seekto(dosHeader.e_lfanew);
    //get(&Signature, sizeof(Signature));

    if (Signature != IMAGE_NT_SIGNATURE) {
        ErrorHandler("Not a Win32 PE or DLL image");
        free(pST);
        free(oST);
        free(pFixups);
        if (!bRamFile) {
            fclose(fExe);
        }
        return NULL;    //failure
    }

    //////////////////   PE HEADER   /////////////////////

    //get(&peHeader, sizeof(peHeader));


    //////////////////   PE OPTIONAL HEADER   /////////////////////

    //SectTableOffset = currentFilePos + peHeader.SizeOfOptionalHeader;
    //get(&optHeader, sizeof(optHeader));

	if (X64GEN) {													//DO THIS EARLIER???
		ExpectedOptHeaderMagic = IMAGE_NT_OPTIONAL_HDR64_MAGIC;
	}
	else {
		ExpectedOptHeaderMagic = IMAGE_NT_OPTIONAL_HDR32_MAGIC;
	}

#if 1	//From AfhPED 1.9
    ///////  Data Directory ///////
	if (!pDD) {
		ErrorHandler("Data Directory not found in optional header");
        free(pST);
        free(oST);
        free(pFixups);
        if (!bRamFile) {
            fclose(fExe);
        }
        return NULL;    //failure
	}
    if (NumberOfRvaAndSizes != IMAGE_NUMBEROF_DIRECTORY_ENTRIES) {
		Warn("Non-standard number of Data Directory entries\n"
             "(%d. instead of the expected %d.)\n", NumberOfRvaAndSizes,
             IMAGE_NUMBEROF_DIRECTORY_ENTRIES);
    }
#endif

	if (optHeader.Magic != ExpectedOptHeaderMagic) {
#ifdef _DEBUG
        ErrorOk("Wrong magic number in optional header\r\n"
				"Expected 0x%x, found 0x%x",
				ExpectedOptHeaderMagic,
				optHeader.Magic);
#else
        ErrorHandler("Wrong magic number in optional header");
#endif
        free(pST);
        free(oST);
        free(pFixups);
        if (!bRamFile) {
            fclose(fExe);
        }
        return NULL;    //failure
    }

    ///////////////////////// BASE-RELOC FIXUPS  ///////////////////////////
    /* Must do *before* trying to update resources! */

    {//scope
        /* Scan base-reloc fixups */
        IMAGE_BASE_RELOCATION fixupHeader;
        struct {
            WORD  Offset  : 12;
            WORD  Type    :  4;
        } fixup;
        int bytesLeftInTable;
        int bytesLeftInBlock;

        if (IMAGE_DIRECTORY_ENTRY_BASERELOC < NumberOfRvaAndSizes &&
            pDD[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress != 0 &&
            pDD[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size != 0) {

            seektoRVA(pDD[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress);
            bytesLeftInTable = pDD[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size;

            /* Initialize the fixup info table to record fixup */
            /* information per section...                      */
            if (pFixups) {
                int i;

                for (i = 0; i < nST; ++i) {
                    pFixups[i].nFixupsToThisSection = 0;
                    pFixups[i].nFixupTargetsToThisSection = 0;
                }
            }

            while (bytesLeftInTable >= sizeof(fixupHeader)) {
                get(&fixupHeader, sizeof(fixupHeader));
                bytesLeftInTable -= sizeof(fixupHeader);
                bytesLeftInBlock = fixupHeader.SizeOfBlock - sizeof(fixupHeader);
#if 0
                /* Count number of (blocks of) fixups to this section...    */
                if (pFixups) {
                    int iSect;

                    iSect = findSection(fixupHeader.VirtualAddress);
                    if (iSect >= 0) {
                        /* This section contains a fixup target */
                        pFixups[iSect].nFixupsToThisSection++;
                    }
                }
#endif
                if (bytesLeftInBlock <= bytesLeftInTable) {
                    while (bytesLeftInBlock >= sizeof(fixup)) {
						UINT64 target64;
                        DWORD target;
                        DWORD rva;
                        size_t saveFilePos;

                        get(&fixup, sizeof(fixup));
                        bytesLeftInBlock -= sizeof(fixup);
                        bytesLeftInTable -= sizeof(fixup);
                        switch (fixup.Type) {
                        case IMAGE_REL_BASED_ABSOLUTE:
							//Fixup is a NOP
                            break;
                        case IMAGE_REL_BASED_HIGHLOW:
                            rva = fixupHeader.VirtualAddress + fixup.Offset;
                            saveFilePos = currentFilePos;
                            seektoRVA(rva);
                            get(&target, sizeof(target));
#if 1   //slower than just counting blocks!!!
                            /* Count number of fixups to this section...    */
                            if (pFixups) {
                                int iSect = findSection(rva);
                                if (iSect >= 0) {
                                    /* This section contains a fixup target */
                                    pFixups[iSect].nFixupsToThisSection++;
                                }
                                else {
                                    Warn("Bad fixup target to RVA 0x%8.8x", rva);
                                }
                            }
#endif
                            if (pFixups) {
                                int iSect = findSection(target - ImageBase);
                                if (iSect >= 0) {
                                    /* Fixup target points into this section */
                                    pFixups[iSect].nFixupTargetsToThisSection++;
                                }
                                else if (target - ImageBase != 0) { //ignore when RVA==0??
                                    Warn("Bad fixup targets RVA 0x%8.8x",
                                         target - ImageBase);
                                }
                            }
                            seekto(saveFilePos);
                            break;
#if 1	//From AfhPED 1.9
						case IMAGE_REL_BASED_DIR64:
                            rva = fixupHeader.VirtualAddress + fixup.Offset;
                            saveFilePos = currentFilePos;
                            seektoRVA(rva);
                            get(&target64, sizeof(target64));
#if 1   //slower than just counting blocks!!!
                            /* Count number of fixups to this section...    */
                            if (pFixups) {
                                int iSect = findSection(rva);
                                if (iSect >= 0) {
                                    /* This section contains a fixup target */
                                    pFixups[iSect].nFixupsToThisSection++;
                                }
                                else {
                                    Warn("Bad fixup target to RVA 0x%8.8x\n", rva);
                                }
                            }
#endif
                            if (pFixups) {
                                int iSect = findSection(target64 - ImageBase);
                                if (iSect >= 0) {
                                    /* Fixup target points into this section */
                                    pFixups[iSect].nFixupTargetsToThisSection++;
                                }
                                else {
                                    Warn("Bad fixup targets RVA 0x%.8I64x\n",
                                         target64 - ImageBase);
                                }
                            }
                            seekto(saveFilePos);
							break;
#endif //AfhPED 1.9
                        case IMAGE_REL_BASED_HIGH:
                        case IMAGE_REL_BASED_LOW:
                        case IMAGE_REL_BASED_HIGHADJ:
                        case IMAGE_REL_BASED_MIPS_JMPADDR:
						case IMAGE_REL_BASED_IA64_IMM64:
                            Warn("Unexpected fixup type (%u.)", fixup.Type);	//ADD MORE SUCH INFO!!!
                            break;
                        default:
                            Warn("Unknown fixup type (%u.)", fixup.Type);
                            break;
                        }//switch
                    }//while
                    if (bytesLeftInBlock != 0) {
                        Warn("Invalid fixup block format (extra bytes)");
                    }
                }
                else {
                    Warn("Invalid block format (not enough bytes)");
                    bytesLeftInTable = 0;   //terminate the scan
                }
            } //while
            if (bytesLeftInTable != 0) {
                Warn("Invalid fixup table format (%d. extra bytes)",
					 bytesLeftInTable);
            }
            if (pST) {
                int i;

                /* Compute section number for presumed .rsrc section... */
                rsrcSect =
                    findSection(pDD[IMAGE_DIRECTORY_ENTRY_RESOURCE].VirtualAddress);
#ifdef VERBOSE
                Info("Presumed .rsrc section number is %d.", rsrcSect);
#endif
                if (rsrcSect < 0) rsrcSect = 9999;  //force comparison okay below...

#ifdef VERBOSE
                Info("The following sections contain base-reloc fixup targets:");
#endif
                for (i = 0; i < nST; ++i) {
                    if (pFixups[i].nFixupsToThisSection != 0) {
#ifdef VERBOSE
                        Info("%7d fixup targets in section %d: %.8s",
                            pFixups[i].nFixupsToThisSection, i, pST[i].Name);
#endif
                        if (i >= rsrcSect) {
#ifdef VERBOSE
                            Warn("Fixups to sections beyond .rsrc!!!");
#endif
                            bFixupsBeyondRsrc = TRUE;
                        }
                    }
                }//for
#ifdef VERBOSE
                Info("Fixup targets point into the following sections:");
#endif
                for (i = 0; i < nST; ++i) {
                    if (pFixups[i].nFixupTargetsToThisSection != 0) {
#ifdef VERBOSE
                        Info("%7d fixup targets to section %d: %.8s",
                             pFixups[i].nFixupTargetsToThisSection, i, pST[i].Name);
#endif
                        if (i >= rsrcSect) {
#ifdef VERBOSE
                            Warn("Fixups to sections beyond .rsrc!!!");
#endif
                            bFixupsBeyondRsrc = TRUE;
                        }
                    }
                }//for
            }//if
        }//if
    }//scope

    //////////////////////////   RESOURCE DIRECTORY   ////////////////////////////
    /* Must do *after* checking fixups! */

    if (IMAGE_DIRECTORY_ENTRY_RESOURCE < NumberOfRvaAndSizes &&
        pDD[IMAGE_DIRECTORY_ENTRY_RESOURCE].VirtualAddress != 0 &&
        pDD[IMAGE_DIRECTORY_ENTRY_RESOURCE].Size != 0) {

        seektoRVA(pDD[IMAGE_DIRECTORY_ENTRY_RESOURCE].VirtualAddress);
        ResRootOffset = currentFilePos;
        NextDirNodeOffset = 0;
        NextDataNodeOffset = 0;
        NextNameNodeOffset = 0;
        NextActualDataOffset = 0;
        pNameNodeHashList = NULL;

        pResDirRoot = DumpResDir(0);

    }
    else {
        ErrorOk("No resources found in template");
        free(pST);
        free(oST);
        free(pFixups);
        if (!bRamFile) {
            fclose(fExe);
        }
        return NULL;    //failure
    }

    return pResDirRoot;
} //afhBeginUpdateResource


BOOL afhEndUpdateResource(
    HANDLE hUpdate,	// update-file handle  
    BOOL fDiscard 	// write flag
   )
{
    ////////////  Write out a new EXE/DLL file   ///////////////////////

    if (fDiscard) {
        // Abandon the update...
        if (!bRamFile) {
            fclose(fExe);
        }
    } else {
        time_t newTime;
        int i;
        int nEntries;

        assignResources(pResDirRoot);

        // WARN if fixups to .rsrc or beyond!!!!
        if (bFixupsBeyondRsrc) {
            Warn("Fixups exist to sections beyond .rsrc;\n"
                 "Output DLL may not be useable!!");
        }

        /* Make sure we have all the data structures we need... */
        if (pDosStub && pST && oST) {
            //okay
        }
        else {
            ErrorHandler("Can't write DLL -- missing data structures");
            if (!bRamFile) {
                fclose(fExe);
            }
            goto cleanup;
        }

        // Make adjustments to PE header...

        time(&newTime);
        peHeader.TimeDateStamp = (DWORD)newTime;   //Time linked/modified

        // Make adjustments to the optional header for the output file...

        /* Linker version = afhped version */
        newOptHeader.MajorLinkerVersion = 1;
        newOptHeader.MinorLinkerVersion = 99;

        /* Adjust size of initialized data */
        newOptHeader.SizeOfInitializedData =
           optHeader.SizeOfInitializedData + outputRVAdelta; //or outputFileOffsetDelta???

        /* Adjust size of image */
        newOptHeader.SizeOfImage =
           optHeader.SizeOfImage + outputRVAdelta; //or outputFileOffsetDelta???

        /* Adjust RVA's in PE optional header */
        newOptHeader.AddressOfEntryPoint = relocateOutputRVA(optHeader.AddressOfEntryPoint); //X64 OK
        newOptHeader.BaseOfCode = relocateOutputRVA(optHeader.BaseOfCode); //X64 OK
		if (!PE32PLUS) {
			//PE32+ format has no BaseOfData field
			newOptHeader.BaseOfData = relocateOutputRVA(optHeader.BaseOfData);
		}

        /* Adjust RVA's and sizes in the output Data Directory */
        oDD[IMAGE_DIRECTORY_ENTRY_RESOURCE].Size = NewResourceSize;
        nEntries = min(NumberOfRvaAndSizes, IMAGE_NUMBEROF_DIRECTORY_ENTRIES);
        for (i = 0; i < nEntries; ++i) {
            oDD[i].VirtualAddress = relocateOutputRVA(pDD[i].VirtualAddress);
        } //for

        if (bRamFile) {
            szOutFileName = szFileName;
        }
        else {
            szOutFileName = tempFileName;
        }
        fOut = fopen(szOutFileName, "wb");
        if (!fOut) {
            ErrorOk("Cannot create file %s", szOutFileName);
            if (!bRamFile) {
                fclose(fExe);
            }
            goto cleanup;
        }

        put(&dosHeader, sizeof(dosHeader));
        if (pDosStub) {
            //write out DOS stub program
            oseekto(dosHeader.e_lfarlc);    //zero-fill?
            put(pDosStub, stubSize);
        }
        else {
            ErrorHandler("No DOS stub");
            if (!bRamFile) {
                fclose(fExe);
            }
            fclose(fOut);
            remove(szOutFileName);
            goto cleanup;
        }
        oseekto(dosHeader.e_lfanew);    //zero-fill?
        put(&Signature, sizeof(Signature));
        put(&peHeader, sizeof(peHeader));
        put(&newOptHeader, peHeader.SizeOfOptionalHeader);
        //seekto(SectTableOffset); ??? needed ???
        put(oST, sizeof(IMAGE_SECTION_HEADER)*nST);

        /* Write out each Section... */
        for (i = 0; i < nST; ++i) {
            if (i == rsrcSect) {
                /* Write new resources... */
                writeResources(pResDirRoot);
            }
            else {
                void *p;
                p = alloc(pST[i].SizeOfRawData);
                seekto(pST[i].PointerToRawData);
                get(p, pST[i].SizeOfRawData);
                oseekto(oST[i].PointerToRawData);
                //should check that pST[i].SizeOfRawData == oST[i].SizeOfRawData ???
                assert(pST[i].SizeOfRawData == oST[i].SizeOfRawData);
                put(p, oST[i].SizeOfRawData);
                //zero-fill to file align if necessary???
                free(p);
            }
        }//for

        /* Write out debug raw data (it's NOT in a Section!) */
        if (IMAGE_DIRECTORY_ENTRY_DEBUG < NumberOfRvaAndSizes &&
            pDD[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress != 0 &&
            pDD[IMAGE_DIRECTORY_ENTRY_DEBUG].Size != 0)
		{
            int nEntries = pDD[IMAGE_DIRECTORY_ENTRY_DEBUG].Size / sizeof(debugDirEntry);
            seektoRVA(pDD[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress);
            for (i = 0; i < nEntries; ++i) {
                get(&debugDirEntry, sizeof(debugDirEntry));
                if (debugDirEntry.PointerToRawData != 0) {
                    size_t saveFilePos = currentFilePos;
                    void *p;

                    p = alloc(debugDirEntry.SizeOfData);
                    seekto(debugDirEntry.PointerToRawData);
                    get(p, debugDirEntry.SizeOfData);
                    //Following needs to be relocated for output file!!!!
                    oseekto(debugDirEntry.PointerToRawData);
                    put(p, debugDirEntry.SizeOfData);
                    free(p);
                    seekto(saveFilePos);
                }//if
            }//for
        }//if

        if (fclose(fOut)) {
            ErrorOk("Error while trying to close file %s", szOutFileName);
        }

        if (!bRamFile) {
            // Close and delete the input file; rename the temporary
            // output file to the input file name.
            if (fclose(fExe)) {
                ErrorOk("Error while trying to close file %s", szFileName);
            }

            if (remove(szFileName)) {
                ErrorOk("Could not delete file %s", szFileName);
            } else {
                if (rename(szOutFileName, szFileName)) {
                    ErrorOk("Could not rename file %s to %s", szOutFileName, szFileName);
                }
            }
        }
    }//if !fDiscard

cleanup:

    //free all mallocs...
    if (pST) {
        free(pST);
        pST = NULL;
    }

    if (oST) {
        free(oST);
        oST = NULL;
    }

    if (pFixups) {
        free(pFixups);
        pFixups = NULL;
    }

    if (pDosStub) {
        free(pDosStub);
        pDosStub = NULL;
    }

    //delete resource tree
    deleteResNodes();
    return TRUE;    //success
} //afhEndUpdateResource
