#ifndef FMDIB_BITMAP_H
#define FMDIB_BITMAP_H

/*
 *   fmdib_bitmap.h
 *
 *   Copyright 1997-2014 AFH Systems
 *
 *   FilterMeister is free software. You can redistribute and modify it
 *   under the terms of the MPL2 License.
 *   See: https://www.mozilla.org/en-US/MPL/2.0/
 *   Any changes to the FilterMeister files have to be published or sent
 *   to harry@thepluginsite.com. The filter plugin code itself does not
 *   need to be disclosed.
 *
 */

//
//	09AUG2013	A. Hunter	1. Changes for Win64
//

#include "fmdib_core.h"

typedef HRESULT (STDAPICALLTYPE *LPTRANSPARENTBLT) (HDC, int, int, int, int, HDC, int, int, int, int, UINT);
LPTRANSPARENTBLT lpTransparentBlt = NULL;	//THREAD-SAFE???

/*
static int fm_setBitmapTile(int x, int y, int iName, int tiles_x, int tiles_y, int tile)
static int fm_setBitmapTileStretch(int x, int y, int iName, int tiles_x, int tiles_y, int tile, int width, int height)
static int fm_setBitmapTileTransparent(int x, int y, int iName, int tiles_x, int tiles_y, int tile, UINT transparent)
static int fm_setBitmapTileStretchTransparent(int x, int y, int iName, int tiles_x, int tiles_y, int tile, int width, int height, UINT transparent)
static int fm_setBitmapTransparent(int x, int y, int iName, UINT transparent)
static int fm_setBitmapStretchTransparent(int x, int y, int iName, int width, int height, UINT transparent)

  BOOL TransparentBlt(
  HDC hdcDest,        // handle to destination DC
  int nXOriginDest,   // x-coord of destination upper-left corner
  int nYOriginDest,   // y-coord of destination upper-left corner
  int nWidthDest,     // width of destination rectangle
  int hHeightDest,    // height of destination rectangle
  HDC hdcSrc,         // handle to source DC
  int nXOriginSrc,    // x-coord of source upper-left corner
  int nYOriginSrc,    // y-coord of source upper-left corner
  int nWidthSrc,      // width of source rectangle
  int nHeightSrc,     // height of source rectangle
  UINT crTransparent  // color to make transparent
); 

Parameters
hdcDest 
[in] Handle to the destination device context. 
nXOriginDest 
[in] Specifies the x-coordinate, in logical units, of the upper-left corner of the destination rectangle. 
nYOriginDest 
[in] Specifies the y-coordinate, in logical units, of the upper-left corner of the destination rectangle. 
nWidthDest 
[in] Specifies the width, in logical units, of the destination rectangle. 
hHeightDest 
[in] Handle to the height, in logical units, of the destination rectangle. 
hdcSrc 
[in] Handle to the source device context. 
nXOriginSrc 
[in] Specifies the x-coordinate, in logical units, of the source rectangle. 
nYOriginSrc 
[in] Specifies the y-coordinate, in logical units, of the source rectangle. 
nWidthSrc 
[in] Specifies the width, in logical units, of the source rectangle. 
nHeightSrc 
[in] Specifies the height, in logical units, of the source rectangle. 
crTransparent 
[in] The RGB color in the source bitmap to treat as transparent. 


stretch mode?!*/

static HBITMAP GDI_CreateHBitmapFromImage(INT_PTR iName) {
	HBITMAP hBmp;

	if (iName == 0) {
		return NULL;
	}

    // First try loading the bitmap from an embedded BITMAP resource...
    hBmp = LoadImage(hDllInstance,
                     (char *)iName,
                     IMAGE_BITMAP,
                     0, 0,
                     LR_DEFAULTCOLOR);

    // If it's not embedded, try searching for it as an external bitmap file...
    if (!hBmp) {
        hBmp = LoadImage(hDllInstance,
                         FmSearchFile((char *)iName),
                         IMAGE_BITMAP,
                         0, 0,
                         LR_LOADFROMFILE);
    }

    if (!hBmp) {
        ErrorOk("Could not load bitmap %s", (char *)iName);
        return NULL;
    }

	return hBmp;
}

//---------------------------------------------------------------------------

static INT_PTR fm_setBitmap(INT_PTR x, INT_PTR y, INT_PTR iName)
{
	int result = false;
	
	HBITMAP hBmp = GDI_CreateHBitmapFromImage(iName);

    if (hBmp) {
		HDC hDC, hMemDC;
		BITMAP bm;
		HBITMAP hOldBitmap;

		hDC = modeGDI? hdcGDI : hdcDIB;

		GetObject(hBmp, sizeof(BITMAP), &bm);
		hMemDC		= CreateCompatibleDC(hDC);
		hOldBitmap	= (HBITMAP)SelectObject(hMemDC, hBmp);
		
		result = BitBlt(hDC, (int)x, (int)y, bm.bmWidth, bm.bmHeight, hMemDC, 0, 0, SRCCOPY);

		SelectObject(hMemDC, hOldBitmap);
		DeleteObject(hBmp);
	}

	return result;
};

static INT_PTR fm_setBitmapTile(INT_PTR x, INT_PTR y, INT_PTR iName,
								INT_PTR tileWidth, INT_PTR tileHeight, INT_PTR tileIndex)
{
	int result = false;

	if (tileWidth > 0 && tileHeight > 0) {
		HBITMAP hBmp = GDI_CreateHBitmapFromImage(iName);

		if (hBmp) {
			int xSrc, ySrc;
			int columns;

			HDC hDC, hMemDC;
			BITMAP bm;
			HBITMAP hOldBitmap;

			hDC = modeGDI? hdcGDI : hdcDIB;

			GetObject(hBmp, sizeof(BITMAP), &bm);
			hMemDC		= CreateCompatibleDC(hDC);
			hOldBitmap	= (HBITMAP)SelectObject(hMemDC, hBmp);

			columns = (bm.bmWidth / (int)tileWidth);

			xSrc = ((int)tileIndex % columns) * (int)tileWidth;
			ySrc = ((int)tileIndex / columns) * (int)tileHeight;
			
			result = BitBlt(hDC, (int)x, (int)y, (int)tileWidth, (int)tileHeight,
							hMemDC, (int)xSrc, (int)ySrc, SRCCOPY);

			SelectObject(hMemDC, hOldBitmap);
			DeleteObject(hBmp);
		}
	}

	return result;
};

static INT_PTR fm_setBitmapTransparent(INT_PTR x, INT_PTR y, INT_PTR iName, UINT_PTR color)
{
	int result = false;

	HBITMAP hBmp = GDI_CreateHBitmapFromImage(iName);

    if (hBmp) {
		HDC hDC, hMemDC;
		BITMAP bm;
		HBITMAP hOldBitmap;

		hDC = modeGDI? hdcGDI : hdcDIB;

		GetObject(hBmp, sizeof(BITMAP), &bm);
		hMemDC		= CreateCompatibleDC(hDC);
		hOldBitmap	= (HBITMAP)SelectObject(hMemDC, hBmp);
		
		// load lpTransparentBlt function
		if (!lpTransparentBlt) {
			if (!hInst_msimg32) {
				hInst_msimg32 = LoadLibrary("msimg32.dll");	//check for error loading!!!
			}
			lpTransparentBlt = ((LPTRANSPARENTBLT) GetProcAddress(hInst_msimg32, "TransparentBlt"));
		}

		if (lpTransparentBlt) {
		result = lpTransparentBlt(hDC, (int)x, (int)y, (int)bm.bmWidth, (int)bm.bmHeight,
								  hMemDC, 0, 0, (int)bm.bmWidth, (int)bm.bmHeight, (UINT)color);
		}	//else return error???

		SelectObject(hMemDC, hOldBitmap);
		DeleteObject(hBmp);
	}

	return result;
};

static INT_PTR fm_setBitmapStretch(INT_PTR x, INT_PTR y, INT_PTR iName, INT_PTR width, INT_PTR height)
{
	int result = false;
	
	HBITMAP hBmp = GDI_CreateHBitmapFromImage(iName);

    if (hBmp) {
		HDC hDC, hMemDC;
		BITMAP bm;
		HBITMAP hOldBitmap;

		hDC = modeGDI? hdcGDI : hdcDIB;

		GetObject(hBmp, sizeof(BITMAP), &bm);
		hMemDC		= CreateCompatibleDC(hDC);
		hOldBitmap	= (HBITMAP)SelectObject(hMemDC, hBmp);
		
		SetStretchBltMode(hDC, HALFTONE);
		SetBrushOrgEx(hDC, 0, 0, NULL);
		result = StretchBlt(hDC, (int)x, (int)y, (int)width, (int)height,
							hMemDC, 0, 0, bm.bmWidth, bm.bmWidth, SRCCOPY);

		SelectObject(hMemDC, hOldBitmap);
		DeleteObject(hBmp);
	}

	return result;
};

static INT_PTR fm_setBitmapStretchTransparent(INT_PTR x, INT_PTR y, INT_PTR iName,
											  INT_PTR width, INT_PTR height, UINT_PTR color)
{
	int result = false;
	
	HBITMAP hBmp = GDI_CreateHBitmapFromImage(iName);

    if (hBmp) {
		HDC hDC, hMemDC;
		BITMAP bm;
		HBITMAP hOldBitmap;

		hDC = modeGDI? hdcGDI : hdcDIB;

		GetObject(hBmp, sizeof(BITMAP), &bm);
		hMemDC		= CreateCompatibleDC(hDC);
		hOldBitmap	= (HBITMAP)SelectObject(hMemDC, hBmp);
		
		// load lpTransparentBlt function
		if (!lpTransparentBlt) {
			if (!hInst_msimg32) {
				hInst_msimg32 = LoadLibrary("msimg32.dll");	//check error loading???
			}
			lpTransparentBlt = ((LPTRANSPARENTBLT) GetProcAddress(hInst_msimg32, "TransparentBlt"));
		}

		if (lpTransparentBlt) {
		result = lpTransparentBlt(hDC, (int)x, (int)y, (int)width, (int)height,
								  hMemDC, 0, 0, bm.bmWidth, bm.bmHeight, (UINT)color);
		}	//else return error???

		SelectObject(hMemDC, hOldBitmap);
		DeleteObject(hBmp);
	}

	return result;
};

#endif
// FMDIB_BITMAP_H