#ifndef FMDIB_TEXT_H
#define FMDIB_TEXT_H

/*
 *   fmdib_text.h
 *
 *   Copyright 1997-2014 AFH Systems
 *
 *   FilterMeister is free software. You can redistribute and modify it
 *   under the terms of the MPL2 License.
 *   See: https://www.mozilla.org/en-US/MPL/2.0/
 *   Any changes to the FilterMeister files have to be published or sent
 *   to harry@thepluginsite.com. The filter plugin code itself does not
 *   need to be disclosed.
 *
 */

//
//	09AUG2013	A. Hunter	1. Changes for Win64
//

HFONT GDInewFont = NULL;	//THREAD-SAFE???
HFONT GDIoldFont = NULL;

int GDIfontAngle = 0;		//THREAD-SAFE???
int GDIfontSize = 8;

//---------------------------------------------------------------------------

static void GDI_destroyFont()
{
	if (GDInewFont) {
		SelectObject(hdcGDI, GDIoldFont);
		DeleteObject(GDInewFont);
		GDInewFont = NULL;
	}
}

//---------------------------------------------------------------------------

static INT_PTR fm_setFont(INT_PTR size, INT_PTR angle, INT_PTR bold, INT_PTR italic, INT_PTR fontname)
{
	if (hbmGDI) {
		SetBkMode(hdcGDI, TRANSPARENT);
		SetTextAlign(hdcGDI, TA_CENTER);

		GDI_destroyFont();

		// 2008-11-13 mwvdlee: Escapement and orientation should be set to same value.
		GDInewFont = CreateFont((int)size, 0, (int)angle, (int)angle,
								(bold? 700 : 400), (int)italic, false, false,
								ANSI_CHARSET, OUT_DEFAULT_PRECIS,
								CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY, DEFAULT_PITCH,
								(LPCTSTR)fontname);
		GDIoldFont = SelectObject(hdcGDI, GDInewFont);
		GDIfontSize = (int)size;
		GDIfontAngle = (int)angle;

		return true;
	}

    return false;
}

//---------------------------------------------------------------------------

static INT_PTR fm_setText(INT_PTR x, INT_PTR y, UINT_PTR color, INT_PTR alignment, INT_PTR text)
{
 if (hbmDIB && hbmGDI) {
  double d, m;
  char * start, * end;
 
  if (!modeGDI) {
   modeGDI = true;
   BitBlt(hdcGDI, 0, 0, GDIcanvasWidth, GDIcanvasHeight, hdcDIB, 0, 0, SRCCOPY);
  }
 
  SetTextColor(hdcGDI, (COLORREF)color);
  SetTextAlign(hdcGDI, (UINT)alignment);
 
  d = (double)GDIfontAngle * M_PI / 1800.;
  m = 0.;
 
  start = (char *)text;
  end = start;
 
  while (*end != 0) {
   m += (double)GDIfontSize;
 
   end = start;
   while (*end != 0 && *end != '\n' && *end != '\r') {
    ++end;
   }
 
   TextOut(hdcGDI, (int)x + (int)(fm_fr2x(d, m)),
				   (int)y + (int)(fm_fr2y(d, m)),
				   start, (int)(end - start));
 
   if (*end == '\r' && *(end + 1) == '\n') {
    ++end;
   }
   start = end + 1;
  }
 
  return true;
 }
 
 return false;
};
 
//---------------------------------------------------------------------------
 
INT_PTR fm_setTextv(INT_PTR x, INT_PTR y, UINT_PTR color, INT_PTR alignment, INT_PTR lpFmt, ...)
{
	char szBuffer[1024];
	va_list ap;
	// va_start(ap, (LPCTSTR)lpFmt);
	va_start(ap, lpFmt);
	_vsnprintf(szBuffer, sizeof(szBuffer), (LPCTSTR)lpFmt, ap);
	szBuffer[sizeof(szBuffer) - 1] = 0;   //ensure NUL-terminated
	va_end(ap);
 
#ifdef BCB
	return fm_setText(x, y, color, alignment, (INT_PTR)szBuffer);
#else
	return fm_setText(x, y, color, alignment, (INT_PTR)fm_formatString(szBuffer));
#endif
} /* setTextv */

//---------------------------------------------------------------------------

#endif
// FMDIB_TEXT_H